/* brick.h */

#ifndef BRICK_H_
#define BRICK_H_


#define X_SHIFT		5
#define Y_SHIFT		X_SHIFT
#define X_BRICK		(1 << X_SHIFT)
#define Y_BRICK		X_BRICK
#define X_MASK		(X_BRICK - 1)
#define Y_MASK		X_MASK
#define BRICK_MASK	(((unsigned long)(~X_MASK)) >> X_SHIFT)


#define BR_EMPTY	1
#define BR_BAD		2

typedef struct Brick Brick;
struct Brick {
    Brick *next;
    Brick *prev;
    Brick *next_modified;
    Brick *next_touched;
    int touched;
    int modified;
    int flags;
    long x, y;
    char orig [Y_BRICK] [X_BRICK+1];
    char line [Y_BRICK] [X_BRICK+1];
    int touched_line [Y_BRICK];
    int empty_line [Y_BRICK];
};



typedef struct Queue Queue;
struct Queue {
    Queue *next;
    long x, y;
};


#define Abs_brick_x(x) ((unsigned long)((x) & BRICK_MASK))
#define Abs_brick_y(y) ((unsigned long)((y) & BRICK_MASK))
#define Abs_brick_from_x(x) (((unsigned long)(x)) >> X_SHIFT)
#define Abs_brick_from_y(y) (((unsigned long)(y)) >> Y_SHIFT)

#define IsInside(px,py,bx,by,nbx,nby) \
		(   (Abs_brick_x (px-bx) < (unsigned long)(nbx))	\
		 && (Abs_brick_x (bx+nbx-px) <= (unsigned long)(nbx))	\
		 && (Abs_brick_y (py-by) < (unsigned long)(nby))	\
		 && (Abs_brick_y (by+nby-py) <= (unsigned long)(nby)))
		 




extern void init_bricks ();
extern void free_all_bricks ();
extern Brick *new_brick ();
extern Brick *search_brick ();
extern void mark_touchedline_brick ();
extern Brick *get_brick ();
extern void warp ();
extern void flush_all_bricks ();
extern void add_burst_zone ();
extern void read_brick_queue ();
extern void reask_bricks ();


extern Brick *brick_list;
extern Brick *bricks_touched;

extern Brick *current_brick;

#endif	/* BRICK_H_ */
