/* builtin_interf.c */

#include <stdio.h>
#include <stdlib.h>

#include "commands.h"
#include "keymap.h"
#include "scrio.h"
#include "mur.h"
#include "error.h"
#include "brick.h"
#include "toplevel.h"
#include "screen.h"
#include "list.h"
#include "parse.h"
#include "tree.h"
#include "mem.h"
#include "readline.h"
#include "interf.h"
#include "util.h"
#include "engio.h"
#include "builtin_interf.h"

#include "externs.h"

/* ********************************************************** */
/* *************** Commands for the interface *************** */
/* ********************************************************** */

int cmd_exit ()
{
    newmode = MODE_EXIT;
    set_fixed_error (ERR_TOPLEVEL, NULL);
    return 0;
}


int special_cmd_map (l, tree)
    List *l;
    Node *tree;	/* has to be copied */
{
    unsigned char *keys;
    unsigned char *ptr;
    static int *buffer = NULL;
    static int buffer_max = 20;
    int buffer_pos;
    List *ll;
    int i;

    if (l == NULL) {
	set_error ("map: null argument");
	return 1;
    }
    i = 0;
    for (ll=l; ll != NULL; ll = ll->next)
	i += strlen (ll->arg) + 1;		   /* spaces and the last \0 */
    ptr = keys = (unsigned char *) smalloc (command_stack, i);
    for (ll=l; ll != NULL; ll = ll->next) {
	int sz = strlen (ll->arg);
	memcpy (ptr, ll->arg, sz);
	ptr += sz;
	*ptr++ = ' ';
    }
    ptr [-1] = '\0';

    if (buffer == NULL)
	buffer = (int *) mmalloc (buffer_max * sizeof (int));
    buffer_pos = 0;

    while (1) {
	unsigned char *item;
	unsigned char *begitem;
	int meta = 0;
	int ctrl = 0;
	int value;

	while (*keys == ' ' || *keys == '\t' || *keys == '\n')
	    keys++;
	begitem = item = keys;
	while (*keys != ' ' && *keys != '\t' && *keys != '\n' && *keys != 0)
	    keys++;
	if (keys == item) /* end of keys */
	    break;
	if (*keys != '\0')
	    *keys++ = '\0'; /* cut a word */

	/*
	 * Now parse item.
	 */
	if (item[1] == 0)					 /* one char */
	    value = item[0];
	else {
	    if (item[0] == '^') {
		ctrl = 1;
		item++;
	    } else {
		while ((item[0] == 'C' || item[0] == 'M')
		       && item[1] == '-') {
		    switch (item[0]) {
		    case 'C':
			ctrl = 1;
			break;
		    case 'M':
			meta = 1;
			break;
		    }
		    item += 2;
		    if (item[0] == '^' && item[1] != '\0') {
			ctrl = 1;
			item++;
			break;
		    }
		}
	    }
	    if (item[0] == 0) {
		char *mem;
	badarg:
		mem = mmalloc (strlen (begitem) + 30);
		sprintf (mem, "bad argument `%s'", begitem);
		set_error_list ("map", l, mem);
		mfree (mem);
		return 0;
	    }
	    if (item[1] == 0)
		value = item[0];
	    else if (!strcasecmp (item, "esc"))
		value = 27;
	    else if (!strcasecmp (item, "tab"))
		value = 9;
	    else if (   !strcasecmp (item, "return")
		     || !strcasecmp (item, "ret"))
		value = 10;
	    else if (!strcasecmp (item, "space"))
		value = 32;
	    else if (!strcasecmp (item, "delete"))
		value = 127;
	    else if (!strcasecmp (item, "backspace"))
		value = 8;
	    else if (STREQ (item, "b1"))
		value = 0x100;
	    else if (STREQ (item, "b2"))
		value = 0x101;
	    else if (STREQ (item, "b3"))
		value = 0x102;
	    else if (STREQ (item, "B1"))
		value = 0x104;
	    else if (STREQ (item, "B2"))
		value = 0x105;
	    else if (STREQ (item, "B3"))
		value = 0x106;
	    else
		goto badarg;
	}

	if (ctrl) {
	    if (value == '?')
		value = 0x7f;
	    else if (value >= 0x100)
		value |= 0x010;
	    else	
		value &= 0x1f;
	}
	if (meta) {
	    if (value >= 0x100)
		value |= 0x008;
	    else
		value |= 0x80;
	}

	if (buffer_pos + 2 > buffer_max) {
	    buffer_max *= 2;
	    buffer = (int *) mrealloc (buffer, buffer_max * sizeof (int));
	}
	buffer [buffer_pos++] = value;
    }

    if (buffer_pos == 0) {
	set_error_list ("map", l, "empty argument");
	return 1;
    }

    if (tree != NULL) {
	tree = copy_tree (tree);
	add_keymap_entry (the_keymap, buffer, buffer_pos, tree);
    } else {
	delete_keymap_entry (the_keymap, buffer, buffer_pos);
    }
    return 0;
}


/* ******************** Move commands ******************** */


static int c_move (dx, dy, l)
    long dx, dy;
    List *l;
{
    if (l != NULL) {
	long val = atol (l->arg);
	dx *= val;
	dy *= val;
    }
    state->pos_x += dx;
    state->pos_y += dy;
    state->lastmove_x = state->pos_x;
    state->lastmove_y = state->pos_y;
    return 0;
}

int cmd_up (l)
    List *l;
{
    return c_move (0, -1, l);
}

int cmd_dn (l)
    List *l;
{
    return c_move (0, 1, l);
}

int cmd_lt (l)
    List *l;
{
    return c_move (-1, 0, l);
}

int cmd_rt (l)
    List *l;
{
    return c_move (1, 0, l);
}


static int c_move_page_float (dx, dy, l)
    long dx, dy;
    List *l;
{
    if (l != NULL) {
	extern double atof ();
	double val = atof (l->arg);
	dx *= val;
	dy *= val;
    }
    state->pos_x += dx;
    state->orig_x += dx;
    state->pos_y += dy;
    state->orig_y += dy;
    state->lastmove_x = state->pos_x;
    state->lastmove_y = state->pos_y;
    return 0;
}

int cmd_pgup (l)
    List *l;
{
    return c_move_page_float (0, -lines, l);
}

int cmd_pgdn (l)
    List *l;
{
    return c_move_page_float (0, lines, l);
}

int cmd_pglt (l)
    List *l;
{
    return c_move_page_float (-cols, 0, l);
}

int cmd_pgrt (l)
    List *l;
{
    return c_move_page_float (cols, 0, l);
}


int cmd_scrollup (l)
    List *l;
{
    return c_move_page_float (0, -1, l);
}

int cmd_scrolldn (l)
    List *l;
{
    return c_move_page_float (0, 1, l);
}

int cmd_scrolllt (l)
    List *l;
{
    return c_move_page_float (-1, 0, l);
}

int cmd_scrollrt (l)
    List *l;
{
    return c_move_page_float (1, 0, l);
}



int cmd_maxup ()
{
    state->pos_y = state->orig_y;
    state->lastmove_x = state->pos_x;
    state->lastmove_y = state->pos_y;
    return 0;
}

int cmd_maxdn ()
{
    state->pos_y = state->orig_y + lines - 1;
    state->lastmove_x = state->pos_x;
    state->lastmove_y = state->pos_y;
    return 0;
}

int cmd_maxlt ()
{
    state->pos_x = state->orig_x;
    state->lastmove_x = state->pos_x;
    state->lastmove_y = state->pos_y;
    return 0;
}

int cmd_maxrt ()
{
    state->pos_x = state->orig_x + cols - 1;
    state->lastmove_x = state->pos_x;
    state->lastmove_y = state->pos_y;
    return 0;
}


int cmd_return (l) /* ret, not return */
    List *l;
{
    long val;
    if (l == NULL)
	val = 1;
    else
	val = atol (l->arg);
    state->pos_x = state->lastmove_x;
    state->pos_y += val;
    return 0;
}


int cmd_tab (l)
    List *l;
{
    long val;
    if (l == NULL)
	val = 1;
    else
	val = atol (l->arg);
    state->pos_x = state->lastmove_x + ((state->pos_x - state->lastmove_x
					 + 8*val) & ~7);
    return 0;
}


/* ******************** Typing things ******************** */


static void c_insert (string,len)
    unsigned char *string;
    long len;
{
    long x, y;
    long bx, by, dx, dy;

    x = state->pos_x;
    y = state->pos_y;
    state->pos_x += len;
    bx = Abs_brick_from_x (x);
    by = Abs_brick_from_y (y);
    dx = x & X_MASK;
    dy = y & Y_MASK;
    while (len) {
	long i;
	unsigned char *ptr;
	long nb = X_BRICK - dx;
	if (nb > len)
	    nb = len;
	if (   current_brick == NULL
	    || current_brick->x != bx
	    || current_brick->y != by)
	    current_brick = get_brick (bx, by);
	ptr = (unsigned char *) current_brick->line[dy]+dx;
	for (i=0; i<nb; i++) {
	    unsigned char c = *string++;
	    if (c < ' ')
		c = ' ';
	    else if (c == 0x7f) {
		ptr++;
		continue;
	    }
	    *ptr++ = c;
	}
	mark_touchedline_brick (current_brick, dy);
	len -= nb;
	bx++;
	dx = 0;
    }
    return;
}


static void c_insertlist (l,separator,lastsep)
    List *l;
    unsigned char separator;
    int lastsep;
{
    List *ll;
    unsigned char *ptr, *string;
    long sz;

    if (l == NULL)
	return;
    sz = 0;
    for (ll=l; ll != NULL; ll = ll->next)
	sz += strlen (ll->arg) + 1;
    ptr = string = (unsigned char *) mmalloc (sz);
    for (; l != NULL; l = l->next) {
	long i = strlen (l->arg);
	memcpy (ptr, l->arg, i);
	ptr += i;
	*ptr++ = separator;
    }
    ptr = string;
    if (!lastsep)
	sz--;
    while (sz) {
	unsigned char c;
	unsigned char *orig = ptr;
	while (sz && (c = *ptr) != '\n') {
	    ptr++;
	    sz--;
	}
	if (ptr != orig)
	    c_insert (orig, ptr-orig);
	if (!sz)
	    break;
	ptr++;
	sz--;
	state->pos_x = state->lastmove_x;
	state->pos_y++;
    }
    mfree (string);
    return;
}


int cmd_insert (l)
    List *l;
{
    c_insertlist (l, ' ', 0);
    return 0;
}

int cmd_paste (l)
    List *l;
{
    long x = state->pos_x;
    long y = state->pos_y;
    long lx = state->lastmove_x;
    long ly = state->lastmove_y;
    c_insertlist (l, '\n', 0);
    state->pos_x = x;
    state->pos_y = y;
    state->lastmove_x = lx;
    state->lastmove_y = ly;
    return 0;
}



int cmd_backspace (l)
    List *l;
{
    long val;
    long i;
    static char *blanks_32 = "                                ";

    if (l == NULL)
	val = 1;
    else {
	val = atol (l->arg);
	if (val <= 0)
	    return 0;
    }
    state->pos_x -= val;
    i = val;
    while (i) {
	if (i >= 32) {
	    c_insert (blanks_32, 32);
	    i -= 32;
	} else {
	    c_insert (blanks_32, i);
	    break;
	}
    }
    state->pos_x -= val;
    return 0;
}

int cmd_grab (l)
    List *l;
{
    long w, h;
    long x, y;
    long bx, by, dx, dy, nbx, nby;
    long last_height;
    long first_width;
    long last_width;
    long width, height;
    long posx, posy;
    long i, j, jj;
    char *mem;
    Brick *br;
    int strip = 0;

    while (l != NULL) {
	if (STREQ (l->arg, "--")) {
	    l = l->next;
	    goto doit;
	}
	if (STREQ (l->arg, "-s")) {
	    l = l->next;
	    strip = 1;
	    continue;
	}
	break;
    }
 doit:
    if (l == NULL
	|| l->next == NULL
	|| l->next->next != NULL) {
	set_error_list ("grab", l, "bad number of arguments");
	return 0;
    }
    if (long_from_string (l->arg, &w)
	|| long_from_string (l->next->arg, &h)
	|| w < 0
	|| h < 0) {
	set_error_list ("grab", l, "bad arguments");
	return 0;
    }
    if (h == 0)
	return 0;
    if (w == 0) {
	while (h--)
	    print_string ("\n");
	return 0;
    }
    mem = mmalloc ((w+1)*h+1);
    for (i=0; i<h; i++)
	mem[(w+1)*(i+1)-1] = '\n';
    mem[(w+1)*h] = '\0';
    x = state->pos_x;
    y = state->pos_y;
    bx = Abs_brick_from_x (x);
    by = Abs_brick_from_y (y);
    dx = x & X_MASK;
    dy = y & Y_MASK;
    nbx = Abs_brick_x (((x+w-1) >> X_SHIFT) - bx + 1);
    nby = Abs_brick_y (((y+h-1) >> Y_SHIFT) - by + 1);
    add_burst_zone (bx, by, nbx, nby);
    read_brick_queue ();
    last_height = (nby == 1) ? h : (((y+h-1) & Y_MASK) + 1);
    first_width = X_BRICK-dx;
    last_width = (nbx == 1) ? w : (((x+w-1) & X_MASK) + 1);
    posy = dy;
    height = Y_BRICK-dy;
    for (j=0; j<nby; j++) {
	long scry = j*Y_BRICK-dy+posy;
	if (j == nby-1) height = last_height;
	posx = dx;
	width = first_width;
	for (i=0; i<nbx; i++) {
	    long scrx = i*X_BRICK-dx+posx;
	    if (i == nbx-1) width = last_width;
	    br = get_brick (bx+i, by+j);
	    for (jj=0; jj<height; jj++) {
		memcpy (mem + (w+1)*(scry+jj)+scrx,
			br->line [posy + jj] + posx,
			width);
	    }
	    posx = 0;
	    width = X_BRICK;
	}
	posy = 0;
	    height = Y_BRICK;
    }
    if (strip) {
	char *src = mem;
	char *dest = mem;
	for (j=h; j; j--) {
	    char *scan = src + w; /* on \n */
	    long i;
	    for (i = w; i && *--scan == ' '; i--)
		;
	    if (dest != src && i != 0)
		bcopy (src, dest, i);
	    src += w+1;
	    dest += i;
	    *dest++ = '\n';
	}
	*dest = '\0';
    }
    print_string (mem);
    mfree (mem);
    return 0;
}



int cmd_putchar (l)
    List *l;
{
    long ofsx, ofsy;
    long x, y;
    long bx, by, dx, dy;
    unsigned char c;

    if (l == NULL
	|| (l->next != NULL
	    && (l->next->next == NULL
		|| l->next->next->next != NULL))) {
	set_error_list ("putchar", l, "bad number of arguments");
	return 0;
    }

    c = l->arg[0];
    if (c == '\0') {
	set_fixed_error (ERR_EXEC, "putchar: empty argument");
	return 0;
    }
    if (c < ' ')
	c = ' ';
    else if (c == 0x7f) /* write a hole */
	return 0;

    if (l->next != NULL) {
	if (long_from_string (l->next->arg, &ofsx)
	    || long_from_string (l->next->next->arg, &ofsy)) {
	    set_error_list ("putchar", l, "bad value");
	    return 0;
	}
    } else
	ofsx = ofsy = 0;
    
    x = state->pos_x + ofsx;
    y = state->pos_y + ofsy;
    bx = Abs_brick_from_x (x);
    by = Abs_brick_from_y (y);
    dx = x & X_MASK;
    dy = y & Y_MASK;
    if (   current_brick == NULL
	|| current_brick->x != bx
	|| current_brick->y != by)
	current_brick = get_brick (bx, by);
    current_brick->line [dy][dx] = c;
    mark_touchedline_brick (current_brick, dy);
    return 0;
}

int cmd_getchar (l)
    List *l;
{
    long ofsx, ofsy;
    long x, y;
    long bx, by, dx, dy;
    static char str[] = "z";

    if (l != NULL
	&& (l->next == NULL
	    || l->next->next != NULL)) {
	set_fixed_error (ERR_EXEC, "getchar: bad number of arguments");
	return 1;
    }
    if (l != NULL) {
	if (long_from_string (l->arg, &ofsx)
	    || long_from_string (l->next->arg, &ofsy)) {
	    set_error_list ("getchar", l, "bad value");
	    return 0;
	}
    } else
	ofsx = ofsy = 0;
    x = state->pos_x + ofsx;
    y = state->pos_y + ofsy;
    bx = Abs_brick_from_x (x);
    by = Abs_brick_from_y (y);
    dx = x & X_MASK;
    dy = y & Y_MASK;
    if (   current_brick == NULL
	|| current_brick->x != bx
	|| current_brick->y != by)
	current_brick = get_brick (bx, by);
    str[0] = current_brick->line [dy][dx];
    print_string (str);
    return 0;
}

int cmd_input (l)
    List *l;
{
    char *prompt;
    char *line;

    if (l != NULL && l->next != NULL) {
	set_error_list ("input", l, "bad number of arguments");
	return 0;
    }
    if (l == NULL) {
	prompt = "";
    } else {
	unsigned char *ptr;
	prompt = mstrdup (l->arg);
	ptr = (unsigned char *) prompt;
	while (*ptr) {
	    if (*ptr < ' ' || *ptr == 0x7f)
		*ptr = ' ';
	    ptr++;
	}
    }
    if (mode == MODE_INTERACT || mode == MODE_MINISHELL) {
	readline_limited = 1;
	readline_width = cols;
	limited_readline_x = 0;
	limited_readline_y = minishelllines - 1;
	readline_winch_routine = limited_readline_winch_routine;
	readline_echo_fun = limited_readline_echo;
	readline_setcur_fun = limited_readline_setcur;
	readline_flush_fun = limited_readline_flush;
	readline_begin_fun = limited_readline_begin;
	readline_end_fun = limited_readline_end;
    } else {
	readline_limited = 0;
	readline_echo_fun = readline_stdout_echo;
	readline_setcur_fun = readline_stdout_setcur;	 /* in case I bugged */
	readline_flush_fun = readline_stdout_flush;
	readline_begin_fun = readline_stdout_begin;
	readline_end_fun = readline_stdout_end;
    }
    line = readline (prompt);
    stty_running ();
    if (line != NULL) {
	print_string (line);
	mfree (line);
    }
    if (l != NULL)
	mfree (prompt);
    return 0;
}



/* **************** Misc commands for the interface *************** */

int cmd_message (l)
    List *l;
{
    int i;
    List *ll;
    char *ptr, *string;
    int lastnl = 1;

    while (l != NULL) {
	if (STREQ (l->arg, "--")) {
	    l = l->next;
	    goto doit;
	}
	if (STREQ (l->arg, "-n")) {
	    l = l->next;
	    lastnl = 0;
	    continue;
	}
	break;
    }
 doit:
    i = 0;
    for (ll=l; ll != NULL; ll = ll->next)
	i += strlen (ll->arg) + 1;
    ptr = string = mmalloc (i+1);  /* room for spaces and the last \n and \0 */
    for (; l != NULL; l = l->next) {
	int sz = strlen (l->arg);
	memcpy (ptr, l->arg, sz);
	ptr += sz;
	*ptr++ = ' ';
    }
    if (lastnl) {
	ptr [-1] = '\n';
	ptr++;
    }
    ptr [-1] = '\0';

    show_message (string, 0);
    if (mode == MODE_INTERACT || mode == MODE_MINISHELL)
	wrefresh (menuw);
    mfree (string);
    return 0;
}



int cmd_cls ()
{
    if (in_screen) {
	werase (menuw);
	state->newcoo = 1;
    } else {
    }
    return 0;
}


int cmd_redraw ()					   /* FORCE a redraw */
{
    if (in_screen) {
	reprint_screen (0);
	wrefresh (curscr);
	wrefresh (mainw);
    }
    return 0;
}


int cmd_display ()			     /* lets the user see the screen */
{
    if (in_screen) {
	reprint_screen (0);
	wrefresh (menuw);
	wrefresh (mainw);
    }
    return 0;
}


int cmd_flush ()
{
    flush_all_bricks ();
    free_all_bricks ();
    state->warped = 1;
    return 0;
}



int cmd_find (l)
    List *l;
{
    char *string;
    char fc;
    long len;
    Brick *br;

    if (l == NULL || l->next != NULL) {
	set_error_list ("find", l, "bad number of arguments");
	return 0;
    }
    string = l->arg;
    len = strlen (string);
    if (len == 0) {
	set_fixed_error (ERR_EXEC, "find: null argument");
	return 0;
    }
    fc = string[0];
    if (brick_list->next->next == brick_list) { /* 1 brick or less */
	int i = (len+63)/32;
	for (; i; i--)
	    get_brick (i-1, 0);
    }
    for (br = brick_list->next; br != brick_list; br = br->next) {
	int i, j;
	if (br->flags & BR_BAD)
	    continue;
	for (j=0; j<Y_BRICK; j++) {
	    char *line = br->line[j];
	    for (i=0; i<X_BRICK; i++) {
		int ii;
		int di;
		Brick *b;
		long bx, by;
		char *bline;
		if (line[i] != fc)
		    continue;
		b = br;
		bx = br->x;
		by = br->y;
		di = i;
		bline = line;
		for (ii=1; ii<len; ii++) {
		    if (++di == X_BRICK) {
			b = search_brick (++bx, by);
			if (b == NULL)
			    goto nextpos;
			di = 0;
			bline = b->line[j];
		    }
		    if (bline[di] != string[ii])
			goto nextpos;
		}
		/* found */
		state->pos_x = (br->x << X_SHIFT) + i;
		state->pos_y = (br->y << Y_SHIFT) + j;
		print_string ("1");
		return 0;
	    nextpos:
		;
	    }
	}
    }
    print_string ("0");	/* not found */
    return 0;
}
    


int cmd_dumpkeymap ()
{
    return 0;
}


int cmd_help ()
{
    return 0;
}


int cmd_mapping ()
{
    return 0;
}


int cmd_goto (l)
    List *l;
{
    if (l == NULL
	|| l->next == NULL
	|| l->next->next != NULL) {
	set_error_list ("goto", l, "bad number of arguments");
	return 0;
    }

    for (; l != NULL; l = l->next) {
	char *string = l->arg;
	int rel;
	long val;

	if (string [0] == '.') {
	    rel = 1;
	    val = 0;
	} else {
	    if (string [0] == '+' && string [1] == '+') {
		rel = 1;
		string += 2;
	    } else if (string [0] == '-' && string [1] == '-') {
		rel = -1;
		string += 2;
	    } else
		rel = 0;
	    if (long_from_string (string, &val)) {
		set_error_list ("goto", l, "bad argument");
		return 1;
	    }
	}

	if (l->next != NULL)
	    if (rel)
		state->pos_x += rel*val;
	    else
		state->pos_x = val - state->trans_x;
	else
	    if (rel)
		state->pos_y += rel*val;
	    else
		state->pos_y = val - state->trans_y;
    }
    state->lastmove_x = state->pos_x;
    state->lastmove_y = state->pos_y;
    return 0;
}


int cmd_filter (l)
    List *l;
{
    strmax buf;
    char *mem;
    long len, lenf;
    if (l == NULL || l->next != NULL) {
	set_error_list ("filter", l, "bad number of arguments");
	return 1;
    }
    len = strlen (l->arg);
    lenf = strlen (FILT);
    mem = mmalloc (len + lenf + 10);
    memcpy (mem, FILT, lenf);
    mem [lenf] = ' ';
    memcpy (mem+lenf+1, l->arg, len);
    mem [lenf+1+len] = '\0';
    flush_write_server (mem);
    mfree (mem);
    read_server (buf);
    print_string (buf);
    return 0;
}


int cmd_stats ()
{
    char buf[STR_MAX_SIZE+2];
    flush_write_server (STAT);
    eread_server (buf);
    if (strcmp (STTS, buf))
	panic (buf);
    while (1) {
	long sz;
	read_server (buf);
	sz = strlen (buf);
	if (sz >= 4) {
	    if (!strncmp (buf, ERRO, 4)
		|| !strncmp (buf, FATL, 4))
		panic ("Stat, ---- attendu");
            if (STREQ (buf, FBRK))
		break;
	    buf[sz++] = '\n';
	    buf[sz] = '\0';
	    print_string (buf);
	}
    }
    return 0;
}





int cmd_warp ()
{
    flush_all_bricks ();
    free_all_bricks ();
    current_brick = NULL;
    warp ();
    find_orig_trans ();
    state->orig_x = state->pos_x-cols/2;
    state->orig_y = state->pos_y-lines/2;
    state->lastmove_x = state->pos_x;
    state->lastmove_y = state->pos_y;
    state->newcoo = 1;
    state->warped = 1;
    return 0;
}


/* ********************************************************************* */
/* ********************************************************************* */
/* ********************************************************************* */


int cmd_toplevel ()
{
    set_fixed_error (ERR_TOPLEVEL, NULL);
    return 0;
}


