/* engio.c */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <signal.h>

#include "interf.h"
#include "mur.h"
#include "util.h"
#include "mem.h"
#include "error.h" /* for errno on this braindead mips */
#include "engio.h"

#include "externs.h"




static int fd_in, fd_out;
static int blocking = 0;
static int nonblocking = 1;



char *eread_server();


int get_errno (string)
    char *string;
{
    if (strncmp (string, ERRO, 4))
	return 0;
    return atoi (string+5);
}


char *init_server ()
{
    static int initialized = 0;
    strmax received;
    int pipe_in[2], pipe_out[2];

    if (initialized)
	return NULL;
    initialized = 1;

    if (pipe (pipe_in))
	panic ("Can't create pipe_in");
    if (pipe (pipe_out))
	panic ("Can't create pipe_out");
    fd_in = pipe_in[0];
    fd_out = pipe_out[1];
    switch (fork ())
    {
    case -1:
	panic ("Can't fork the guardian");
    case 0: /* son */
	signal (SIGINT, SIG_IGN);
	close (fd_in);
	close (fd_out);
	pipe_in[1] = fd_safe (pipe_in[1]);
	pipe_out[0] = fd_safe (pipe_out[0]);
	close (0);
	close (1);
	dup2 (pipe_out[0], 0);
	dup2 (pipe_in[1], 1);
	fcntl (0, F_SETFD, 0); /* No close on exec */
	fcntl (1, F_SETFD, 0);
	ioctl (0, FIONBIO, &blocking);
	ioctl (1, FIONBIO, &blocking);
	execl ("/bin/sh", "sh", "-c", engine, NULL);
	panic ("Can't execl the guardian");
    }
    close (pipe_out[0]);
    close (pipe_in[1]);
    ioctl (fd_in, FIONBIO, &blocking);
    ioctl (fd_out, FIONBIO, &nonblocking);

    eread_server (received);
    if (strncmp (received, HELO, 4))
	panic ("I don't understand the guardian");
    if (strlen (received) < 5)
	return "(unknown)";				  /* version unknown */
    else
	return received+5;					  /* version */
}


char *read_server (ptr)
    char *ptr;
{
#define BUF_SIZE STR_MAX_SIZE
    static strmax buf;
    static int len = 0;
    static int pos = 0;
    char *cpy = ptr;
    int cnt = 0;

    while (1) {
	while (pos < len) {
	    unsigned char c = buf[pos++];
	    switch (c) {
	    case '\n':
		*ptr = '\0';
		return cpy;
	    default:
		if (c < ' ')
		    continue;
		if (cnt < STR_MAX_SIZE) {
		    *ptr++ = c;
		    cnt++;
		}
	    }
	}
	len = read (fd_in, buf, BUF_SIZE);
	if (len <= 0) {
	    if (len == 0 || errno != EINTR)
		panic ("The guardian is dead");
	}
	pos = 0;
    }
}


char *eread_server (ptr)
    char *ptr;
{
    read_server (ptr);
    if (!strncmp (ptr, ERRO, 4))
	panic (ptr);
    if (!strncmp (ptr, FATL, 4))
	panic (ptr);
    return ptr;
}




typedef struct Writequeue Writequeue;
struct Writequeue {
    Writequeue *next;
    int len;
    char *string;
};

/*
 * freelist for Writequeue
 */

static Writequeue *Writequeue_freelist = NULL;

static Writequeue *new_Writequeue ()
{
    Writequeue *w;
    if (Writequeue_freelist) {
	w = Writequeue_freelist;
	Writequeue_freelist = w->next;
    } else
	w = New (Writequeue);
    return w;
}

static void free_Writequeue (w)
    Writequeue *w;
{
    w->next = Writequeue_freelist;
    Writequeue_freelist = w;
}



static Writequeue *write_queue_head = NULL;
static Writequeue **write_queue_tailp = &write_queue_head;




int write_server (mesg)
    char *mesg;
{
    if (mesg != NULL) {
	Writequeue *ptr = new_Writequeue ();
	int len = strlen (mesg);
	ptr->next = NULL;
	*write_queue_tailp = ptr;
	write_queue_tailp = &(ptr->next);
	ptr->string = mmalloc (len+1); /* for newline */
	memcpy (ptr->string, mesg, len);
	ptr->string[len] = '\n';
	ptr->len = len+1;
    }
    while (write_queue_head != NULL) {
	Writequeue *ptr;
	int len;
	int res;
	mesg = write_queue_head->string;
	len = write_queue_head->len;
	res = write (fd_out, mesg, len);
	if (res <= 0) {
	    if (res < 0 && (errno == EWOULDBLOCK || errno == EAGAIN))
		return 1;
	    panic ("The guardian is dead");
	}
	if (res != len) {
	    bcopy (mesg+res, mesg, len-res); /* correct if overlap */
	    write_queue_head->len -= res;
	    return 1;
	}
	mfree (mesg);
	ptr = write_queue_head;
	write_queue_head = write_queue_head->next;
	free_Writequeue (ptr);
    }
    write_queue_tailp = &write_queue_head;
    return 0;
}


int flush_once ()
{
    return write_server (NULL);
}


void flush_write_server (mesg)
    char *mesg;
{
    ioctl (fd_out, FIONBIO, &blocking);
    write_server (mesg);
    while (write_server (NULL))
	;
    ioctl (fd_out, FIONBIO, &nonblocking);
}


void exit_server ()
{
    strmax received;

    flush_write_server (EXIT);
    read_server (received);
    if (!strcmp (received, BYE))
	return;
    fprintf (stderr, "%s: Problem with the guardian in quitting\n", myname);
}



#if 0
long nb_to_read ()
{
    long ret;

    if (ioctl (fd_in, FIONREAD, &ret))
	panic ("ioctl impossible");
    return ret;
}
#endif /* 0 */

#if 0
int eof ()
{
    return (nb_to_read () == 0);
}
#endif /* 0 */

