/* error.c */

#include <stdio.h>
#include <errno.h>

#include "mem.h"
#include "list.h"
#include "error.h"

#include "externs.h"



enum Error the_error = ERR_NONE; /* 0 */

static int the_error_allocated = 0;
char *the_error_message = NULL;			       /* mmalloc'ed message */

void no_error ()
{
    if (the_error_allocated) {
	mfree (the_error_message);
	the_error_allocated = 0;
    }
    the_error = ERR_NONE;
}

void set_error (err, s)
    enum Error err;
    char *s;
{
    if (the_error_allocated)
	mfree (the_error_message);
    else
	the_error_allocated = 1;
    the_error = err;
    the_error_message = mstrdup (s);
}

void set_fixed_error (err, s)
    enum Error err;
    char *s;
{
    if (the_error_allocated) {
	mfree (the_error_message);
	the_error_allocated = 0;
    }
    the_error = err;
    the_error_message = s;
}


void set_error_errno (s)
    char *s;
{
    char *mem;
    extern int sys_nerr;
    extern char *sys_errlist[];

    if (errno < sys_nerr) {
	mem = mmalloc (strlen (s) + strlen (sys_errlist[errno]) + 10);
	sprintf (mem, "%s: %s", s, sys_errlist[errno]);
    } else {
	mem = mmalloc (strlen (s) + 40);
	sprintf (mem, "%s: Unknown error ?", s);
    }
    set_error (ERR_EXEC, mem);
    mfree (mem);
}



void set_error_list (fun, l, msg)
    char *fun;
    List *l;
    char *msg;
{
    char *mem;
    int sz = strlen (fun) + strlen (msg) + 10;
    List *ll;
    for (ll=l; ll != NULL; ll = ll->next)
	sz += strlen (ll->arg) + 1;
    mem = mmalloc (sz);
    strcpy (mem, fun);
    for (; l != NULL; l = l->next) {
	strcat (mem, " ");
	strcat (mem, l->arg);
    }
    strcat (mem, ": ");
    strcat (mem, msg);
    set_error (ERR_EXEC, mem);
    mfree (mem);
}


void set_error_list_errno (fun, l)
    char *fun;
    List *l;
{
    char *mem;
    int sz = strlen (fun) + 10;
    List *ll;
    for (ll=l; ll != NULL; ll = ll->next)
	sz += strlen (ll->arg) + 1;
    mem = mmalloc (sz);
    strcpy (mem, fun);
    for (; l != NULL; l = l->next) {
	strcat (mem, " ");
	strcat (mem, l->arg);
    }
    set_error_errno (mem);
    mfree (mem);
}
