/* keymap.c */

#include <stdio.h>
#include <unistd.h>

#include "keymap.h"
#include "tree.h"
#include "list.h"
#include "mem.h"

#include "externs.h"



Keymap *the_keymap = NULL;



typedef struct Keycommand_queue Keycommand_queue;
struct Keycommand_queue {
    struct Keycommand_queue *next;
    Keycommand *kc;
};

static Keycommand_queue *pending_kc = NULL;



int mouse_pos_x = 0;
int mouse_pos_y = 0;


List *getvar_mousepos ()
{
    static char buf1[20];
    static char buf2[20];
    static List list1;
    static List list2;
    sprintf (buf1, "%d", mouse_pos_x);
    sprintf (buf2, "%d", mouse_pos_y);
    list1.arg = buf1;
    list2.arg = buf2;
    list1.next = &list2;
    list2.next = NULL;
    return &list1;
}

void setvar_mousepos (l)
    List *l;
{
    /* read only variable */
}






static void do_free_kc (kc)
    Keycommand *kc;
{
    if (kc != NULL) {
	free_tree (kc->tree);
	mfree (kc);
    }
}

void free_pending_kcs ()
{
    while (pending_kc != NULL) {
	Keycommand_queue *next = pending_kc->next;
	do_free_kc (pending_kc->kc);
	mfree (pending_kc);
	pending_kc = next;
    }
}

static void free_kc (kc)
    Keycommand *kc;
{
    if (kc->inuse) {
	Keycommand_queue *new = New (Keycommand_queue);
	new->kc = kc;
	new->next = pending_kc;
	pending_kc = new;
    } else
	do_free_kc (kc);
}



Keymap *new_keymap ()
{
    Keymap *keymap;
    int c;

    keymap = New (Keymap);
    for (c=0; c<KEYMAP_SIZE; c++) {
	(*keymap)[c].keymap = NULL;
	(*keymap)[c].kc = NULL;
    }
    return keymap;
}


void free_keymap_keymap (keymap, c)
    Keymap *keymap;
    int c;
{
    if ((*keymap)[c].keymap != NULL) {
	free_keymap ((*keymap)[c].keymap);
	(*keymap)[c].keymap = NULL;
    }
}


void free_keymap_kc (keymap, c)
    Keymap *keymap;
    int c;
{
    if ((*keymap)[c].kc != NULL) {
	free_kc ((*keymap)[c].kc);
	(*keymap)[c].kc = NULL;
    }
}


void free_keymap (keymap)
    Keymap *keymap;
{
    int c;

    for (c=0; c<KEYMAP_SIZE; c++) {
	free_keymap_keymap (keymap, c);
	free_keymap_kc (keymap, c);
    }
    mfree (keymap);
}


void add_keymap_entry (keymap, keyseq, len, tree)
    Keymap *keymap;
    int *keyseq;
    int len;
    Node *tree;
{
    int c;

 top:
    c = keyseq[0];
    if (len == 1) {
	Keycommand *kc;
	free_keymap_keymap (keymap, c);
	free_keymap_kc (keymap, c);
	kc = New (Keycommand);
	kc->inuse = 0;
	kc->tree = tree;
	(*keymap)[c].kc = kc;
    } else {
	free_keymap_kc (keymap, c);
	if ((*keymap)[c].keymap == NULL)
	    (*keymap)[c].keymap = new_keymap ();
	/* add_keymap_entry ((*keymap)[c].keymap, keyseq+1, len-1, tree); */
	keymap = (*keymap)[c].keymap;
	keyseq++;
	len--;
	goto top;
    }
}



void delete_keymap_entry (keymap, keyseq, len)
    Keymap *keymap;
    int *keyseq;
    int len;
{
    int c;
    Keymap *km;

 top:
    c = keyseq[0];
    if (len == 1) {
	free_keymap_keymap (keymap, c);
	free_keymap_kc (keymap, c);
    } else {
	free_keymap_kc (keymap, c);
	if ((*keymap)[c].keymap != NULL) {
	    /* delete_keymap_entry ((*keymap)[c].keymap, keyseq+1, len-1); */
	    km = (*keymap)[c].keymap;
	    (*keymap)[c].keymap = NULL;
	    keymap = km;
	    keyseq++;
	    len--;
	    goto top;
	}
    }
}


void init_keymap ()
{
    if (the_keymap == NULL)
	the_keymap = new_keymap ();
}


int walk_keymap (keymap, cc, nextkeymapp, kcp)
    Keymap *keymap;
    int cc;
    Keymap **nextkeymapp;
    Keycommand **kcp;
{
    int c;
    if (cc == -1 || cc == -2) {
	fprintf (stderr, "walk keymap bug : cc = %d \n", cc);
	sleep (1);
	return 0;
    }
    c = cc & 0x1ff;
    if (c >= 0x120) {
	fprintf (stderr, "walk keymap bug : cc = 0x%08x \n", cc);
	sleep (1);
	return 0;
    } else if (c >= 0x100) {
	mouse_pos_x = ((unsigned int)(cc & 0xff000000)) >> 24;
	mouse_pos_y = ((unsigned int)(cc & 0x00ff0000)) >> 16;
    }
    if ((*keymap)[c].keymap != NULL) {
	*nextkeymapp = (*keymap)[c].keymap;
	return 1; /* keymap */
    }
    if ((*keymap)[c].kc != NULL) {
	*kcp = (*keymap)[c].kc;
	return 2; /* keycommand */
    }
    return 0; /* not found */
}

