/* list.c */

#include "mem.h"
#include "error.h"
#include "list.h"
#include "commands.h"

#include "externs.h"




List *snew_List ()
{
    return SNew (command_stack, List);
}


List *scopy_of_list (l)					/* copy in the stack */
    List *l;
{
    List *ll, *r;

    if (l == NULL)
	return NULL;
    r = ll = snew_List ();
    while (1) {
	ll->arg = sstrdup (command_stack, l->arg);
	l = l->next;
	if (l != NULL) {
	    ll->next = snew_List ();
	    ll = ll->next;
	} else {
	    ll->next = NULL;
	    break;
	}
    }
    return r;
}


int nbargs_list (l)
    List *l;
{
    int i;

    for (i=0; l != NULL; l = l->next)
	i++;
    return i;
}



List *append_lists (l1,l2)				/* nothing is copied */
    List *l1,*l2;
{
    List *l;

    if (l1 == NULL)
	return l2;
    if (l2 == NULL)
	return l1;
    l = l1;
    while (l->next != NULL)
	l = l->next;
    l->next = l2;
    return l1;
}


List *sconcat_lists (l1,l2)	   /* In the stack. Destroys previous lists. */
    List *l1,*l2;
{
    int len1, len2;
    
    if (l1 == NULL)
	return l2;
    if (l2 == NULL)
	return l1;
    len1 = nbargs_list (l1);
    if (len1 == 1) {				       /* in-place modify l2 */
	char *w1 = l1->arg;
	int szw1 = strlen (w1);
	List *ll = l2;
	do {
	    char *w2 = ll->arg;
	    int szw2 = strlen (w2);
	    char *nw = smalloc (command_stack, szw1+szw2+1);	 /* in stack */
	    memcpy (nw, w1, szw1);
	    memcpy (nw+szw1, w2, szw2+1); /* with the \0 */
	    ll->arg = nw;
	    ll = ll->next;
	} while (ll != NULL);
	return l2;
    }
    len2 = nbargs_list (l2);
    if (len2 == 1) {				       /* in-place modify l1 */
	char *w2 = l2->arg;
	int szw2 = strlen (w2);
	List *ll = l1;
	do {
	    char *w1 = ll->arg;
	    int szw1 = strlen (w1);
	    char *nw = smalloc (command_stack, szw1+szw2+1);	 /* in stack */
	    memcpy (nw, w1, szw1);
	    memcpy (nw+szw1, w2, szw2+1); /* with the \0 */
	    ll->arg = nw;
	    ll = ll->next;
	} while (ll != NULL);
	return l1;
    }
    if (len1 == len2) {				       /* in-place modify l1 */
	List *ll = l1;
	do {
	    char *w1 = ll->arg;
	    char *w2 = l2->arg;
	    int szw1 = strlen (w1);
	    int szw2 = strlen (w2);
	    char *nw = smalloc (command_stack, szw1+szw2+1);	 /* in stack */
	    memcpy (nw, w1, szw1);
	    memcpy (nw+szw1, w2, szw2+1); /* with the \0 */
	    ll->arg = nw;
	    ll = ll->next;
	    l2 = l2->next;
	} while (ll != NULL);
	return l1;
    }
    set_error (ERR_EXEC, "bad concatenation");
    return NULL;
    /*NOTREACHED*/
}


/*
 * freelist for List (in permanent memory, not in the stack).
 */

static List *List_freelist = NULL;


List *new_List ()
{
    List *l;
    if (List_freelist) {
	l = List_freelist;
	List_freelist = l->next;
    } else
	l = New (List);
    return l;
}

void free_List (l)
    List *l;
{
    l->next = List_freelist;
    List_freelist = l;
}



List *permanent_copy_of_list (l)       /* everything in list copied into mem */
    List *l;
{
    List *ll, *r;

    if (l == NULL)
	return NULL;
    r = ll = new_List ();
    while (1) {
	ll->arg = mstrdup (l->arg);
	l = l->next;
	if (l != NULL) {
	    ll->next = New (List);
	    ll = ll->next;
	} else {
	    ll->next = NULL;
	    break;
	}
    }
    return r;
}


void free_permanent_list (l)
    List *l;
{
    while (l != NULL) {
	List *ll = l->next;
	mfree (l->arg);
	free_List (l);
	l = ll;
    }
}


void print_list (l)
    List *l;
{
    printf ("(");
    for (; l != NULL; l = l->next) {
	printf ("%s", l->arg);
	if (l->next != NULL)
	    printf (" ");
    }
    printf (")");
}
