/* mur.c */

/*
 *	Mur v2.0 - 1993/05/01
 *
 *	Interface by <Florent.Guillaume@ens.fr>
 */


#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/types.h> /* for time_t */

#include "keymap.h"
#include "scrio.h"
#include "engio.h"
#include "toplevel.h"
#include "commands.h"
#include "brick.h"
#include "screen.h"
#include "parse.h"
#include "tree.h"
#include "error.h"
#include "mur.h"
#include "mem.h" /* for rindex */

#include "externs.h"

extern time_t time ();

char *myname;
char *server_ver;
char *engine;


extern void srand48 ();



void panic (s, t1, t2, t3, t4)
     char *s, *t1, *t2, *t3, *t4;
{
    exit_scr ();
    fprintf (stderr, "\n%s: ", myname);
    fprintf (stderr, s, t1, t2, t3, t4);
    fprintf (stderr, "\n");
    stty_original ();
    exit (255);
}




void error (s)
     char *s;
{
    fprintf (stderr, "\n%s: error: %s\n", myname, s);
    sleep (1);
    return;
}


static void catch_children ()
{
    while (waitpid (-1, NULL, WNOHANG) > 0)
	;
}


int main (argc, argv)
     int argc;
     char **argv;
{
    char **commands = NULL;
    int nbcommands = 0;
    int opt_fast_init = 0;


    engine = DEF_ENGINE;
    myname = rindex (*argv, '/');
    if (myname == NULL) myname = *argv; else myname++;
    argc--;
    argv++;

    while (argc > 0) {
	if (!strcmp ("-g", argv[0])) {
	    if (argc < 2)
		goto usage;
	    argv++;
	    argc--;
	    engine = argv[0];
	} else if (!strcmp ("-c", argv[0])) {
	    flag_nocoo = 1;
	} else if (!strcmp ("-q", argv[0])) {
	    opt_fast_init = 1;
	} else if (!strcmp ("-e", argv[0])) {
	    if (argc < 2)
		goto usage;
	    argv++;
	    argc--;
	    if (!nbcommands)
		commands = (char **) mmalloc (sizeof (char *));
	    else
		commands = (char **) mrealloc (commands,
					       (nbcommands+1)*sizeof (char *));
	    commands[nbcommands] = mstrdup (argv[0]);
	    nbcommands++;
	    break;
	} else if (!strcmp ("-h", argv[0])) {
	    goto usage;
	} else if (!strcmp ("-v", argv[0])) {
	    printf ("Mur Interface v%s -- by Florent.Guillaume@ens.fr\n", MUR_VER);
	    exit (0);
	} else {
	    fprintf (stderr, "%s: bad argument `%s'\n", myname, argv[0]);
	usage:
	    fprintf (stderr, "usage: %s [flags]\n", myname);
	    fprintf (stderr,
		     "\t-h             This help message\n");
	    fprintf (stderr,
		     "\t-v             Prints the interface version (%s)\n", MUR_VER);
	    fprintf (stderr,
		     "\t-g <guardian>  Use another guardian\n");
	    fprintf (stderr,
		     "\t-c             Don't show coordinates\n");
	    fprintf (stderr,
		     "\t-q             Don't read any init file\n");
	    fprintf (stderr,
		     "\t-e <commands>  Execute some commands\n");
	    exit (2);
	    break;
	}
	argc--;
	argv++;
    }
    signal (SIGCLD, catch_children);
    signal (SIGPIPE, SIG_IGN);
    srand48 (time (NULL) + getpid ());
    init_tty ();
    init_keys ();
    init_commands ();
    init_keymap ();
    init_bricks ();
    init_specialvars ();
    server_ver = mstrdup (init_server ());
    init_state ();
    if (!opt_fast_init) {
	extern char *getenv ();
	FILE *file;
	char *mem;
	char *home;
	file = fopen (DEFAULTSFILE, "r");
	if (file == NULL) {
	    set_error_errno (DEFAULTSFILE);
	    fprintf (stderr, "%s\n", the_error_message);
	    no_error ();
	    newmode = MODE_SHELL;
	} else {
	    exec_file (file);
	    if (the_error && the_error != ERR_TOPLEVEL) {
		fprintf (stderr, "error in %s\n", DEFAULTSFILE);
		fprintf (stderr, "%s\n", the_error_message);
		no_error ();
		newmode = MODE_SHELL;
	    }
	}
	home = getenv ("HOME");
	if (home == NULL)
	    home = ".";
	mem = mmalloc (strlen (home) + strlen (CONFIGFILE) + 10);
	strcpy (mem, home);
	strcat (mem, "/");
	strcat (mem, CONFIGFILE);
	file = fopen (mem, "r");
	if (file != NULL) {
	    exec_file (file);
	    if (the_error && the_error != ERR_TOPLEVEL) {
		fprintf (stderr, "error in %s\n", mem);
		fprintf (stderr, "%s\n", the_error_message);
		no_error ();
		newmode = MODE_SHELL;
	    }
	}
	mfree (mem);
    }
    if (nbcommands) {
	int i;
	for (i=0; i<nbcommands; i++) {
	    if (!the_error)
		exec_string (commands[i]);
	    mfree (commands[i]);
	}
	mfree (commands);
    }
    toplevel ();
    flush_all_bricks ();
    exit_server ();
    stty_original ();
    exit (0);
}
