/* parse.y */

/* Yacc grammar for mur. */

%{

#include <stdio.h>

/* #include "mur.h" */
/* #include "commands.h" */
#include "lex.h"
#include "tree.h"
#include "error.h"
#include "mem.h"


Stack *parse_stack; /* not declared in .h, imported by lex, tree */



static Node *parse_tree;


%}





%token WORD VARSUB COUNT FLAT FUN MAP IF THEN ELSE WHILE LOCAL FOR
%token RETURN BREAK VOID CATCH
%token LT LE GT GE ANDAND OROR NEQ
%token END
/* HUH is used nowhere, so that it generates an error */
%token HUH

%left '\n' ';'
%left ')'
/* %right '[' */
%left OROR
%left ANDAND
%left '=' NEQ
%left LT LE GT GE 
%left '+' '-'
%left '*' '/'
%right '$'
%left VARSUB



%union {
    Node *node;
    char *word;
}

%type <word>	WORD keyword opword
%type <node>	all line body brace cmd simple fword word sword nlwords
		comword arith 

%start all

%%

/* can be NULL */
all	: line end		{ parse_tree = $1;
				  YYACCEPT; }
	| error end		{ yyerrok;
				  parse_tree = NULL;	
				  YYABORT; }
	;

end	: END /* EOF */		/* nothing */
	| '\n'			/* nothing */
	;
							 /* line can be NULL */
line    : cmd			/* copy */
	| line ';' cmd		{ $$ = mkappend(nCmd,$1,$3); }
	;
							  /* cmd can be NULL */
cmd	: /* empty */		{ $$ = NULL; }
    	| simple		/* copy */
	| brace			/* copy */
	| FUN fword cmd		{ $$ = mknode(nFn);
				  $$->u[0].p = $2;
				  $$->u[1].p = $3; } /* maybe NULL */
	| MAP word cmd		{ $$ = mknode(nMap);
				  $$->u[0].p = $2;
				  $$->u[1].p = $3; } /* maybe NULL */
	| fword '=' word	{ $$ = mknode(nSetvar);
				  $$->u[0].p = $1;
				  $$->u[1].p = $3; }
	| LOCAL fword		{ $$ = mknode(nLocal);
				  $$->u[0].p = $2;
				  $$->u[1].p = NULL; }
	| LOCAL fword '=' word	{ $$ = mknode(nLocal);
				  $$->u[0].p = $2;
				  $$->u[1].p = $4; }
	| IF word brace		{ $$ = mknode(nIf);
				  $$->u[0].p = $2;
				  $$->u[1].p = $3;
			      	  $$->u[2].p = NULL; }
	| IF word THEN cmd ';' ELSE cmd { $$ = mknode(nIf);
					  $$->u[0].p = $2;
					  $$->u[1].p = $4;
					  $$->u[2].p = $7; }
	| IF word THEN brace ELSE brace { $$ = mknode(nIf);
					  $$->u[0].p = $2;
					  $$->u[1].p = $4;
					  $$->u[2].p = $6; }
	| WHILE word cmd	{ $$ = mknode(nWhile);
				  $$->u[0].p = $2;
				  $$->u[1].p = $3; }
	| FOR word word cmd	{ $$ = mknode(nFor);
				  $$->u[0].p = $2;
				  $$->u[1].p = $3;
				  $$->u[2].p = $4; }
	| RETURN		{ $$ = mknode(nReturn); $$->u[0].p = NULL; }
	| RETURN word		{ $$ = mknode(nReturn); $$->u[0].p = $2; }
	| BREAK			{ $$ = mknode(nBreak); $$->u[0].p = NULL; }
	| BREAK word		{ $$ = mknode(nBreak); $$->u[0].p = $2; }
	| VOID cmd		{ $$ = mknode(nVoid); $$->u[0].p = $2; }
	| CATCH cmd		{ $$ = mknode(nCatch); $$->u[0].p = $2; }
	;
							/* brace can be NULL */
brace	: '{' body '}'		{ $$ = mknode(nBrace);
				  $$->u[0].p = $2; }
	;
							 /* body can be NULL */
body	: cmd			/* copy */
        | body ';' cmd		{ $$ = mkappend(nCmd,$1,$3); }
	| body '\n' cmd		{ $$ = mkappend(nCmd,$1,$3); }
	;

simple	: fword			/* copy */
	| simple word		{ $$ = mkappend(nArgs,$1,$2); }
	;

fword	: comword		/* copy */
	| opword		{ $$ = mknode(nWord); $$->u[0].s = $1; }
	| fword '^' sword	{ $$ = mknode(nConcat);
				  $$->u[0].p = $1;
				  $$->u[1].p = $3; }
	;

word	: sword			/* copy */
	| word '^' sword	{ $$ = mknode(nConcat);
				  $$->u[0].p = $1;
				  $$->u[1].p = $3; }
	;

sword	: comword		/* copy */
	| opword		{ $$ = mknode(nWord); $$->u[0].s = $1; }
	| keyword		{ $$ = mknode(nWord); $$->u[0].s = $1; }
	;

nlwords	: /* empty */		{ $$ = NULL; }
	| nlwords '\n'		/* copy */
	| nlwords word		{ $$ = mkappend(nArgs,$1,$2); }
	;

comword	: '$' sword		{ $$ = mknode(nVar); $$->u[0].p = $2; }
	| '$' sword VARSUB nlwords ')' { $$ = mknode(nVarsub);
					 $$->u[0].p = $2;
					 $$->u[1].p = $4; }
	| COUNT sword		{ $$ = mknode(nCount); $$->u[0].p = $2; }
	| FLAT sword		{ $$ = mknode(nFlat); $$->u[0].p = $2; }
	| '`' word brace	{ $$ = mknode(nBackq);
				  $$->u[0].p = $2;
				  $$->u[1].p = $3; }
	| '`' brace		{ $$ = mknode(nBackq);
				  $$->u[0].p = NULL;
				  $$->u[1].p = $2; }
	| '(' nlwords ')'	{ $$ = $2; }
	| '[' arith ']'		{ $$ = mknode(nArith); $$->u[0].p = $2; }
	| WORD			{ $$ = mknode(nWord);
				  $$->u[0].s = $1; }
	;
/*
  if [ $a + 8 * $b = 5 ] then zap; else blop
  fun fact {
    if [$0 < 2] then return 1; else return [$0 * fact [$0 - 1]]
  }



*/

arith	: comword		/* copy */
	| keyword		{ $$ = mknode(nWord); $$->u[0].s = $1; }
	| arith '+' arith	{ $$ = mkappend(nAdd,$1,$3); }
	| arith '-' arith	{ $$ = mkappend(nSub,$1,$3); }
	| arith '*' arith	{ $$ = mkappend(nMul,$1,$3); }
	| arith '/' arith	{ $$ = mkappend(nDiv,$1,$3); }
	| arith '=' arith	{ $$ = mkappend(nEq,$1,$3); }
	| arith NEQ arith	{ $$ = mkappend(nNeq,$1,$3); }
	| arith LT arith	{ $$ = mkappend(nLt,$1,$3); }
	| arith LE arith	{ $$ = mkappend(nLe,$1,$3); }
	| arith GT arith	{ $$ = mkappend(nGt,$1,$3); }
	| arith GE arith	{ $$ = mkappend(nGe,$1,$3); }
	| arith ANDAND arith	{ $$ = mkappend(nAndand,$1,$3); }
	| arith OROR arith	{ $$ = mkappend(nOror,$1,$3); }
	| comword '[' arith ']' { Node *n1 = mknode(nArith);
				  n1->u[0].p = $3;
				  $$ = mknode(nBackq);
				  $$->u[0].p = NULL;
				  $$->u[1].p = mkappend(nArgs,$1,n1); }
	;

keyword	: FUN			{ $$ = "fun"; }
	| MAP			{ $$ = "map"; }
	| IF			{ $$ = "if"; }
	| THEN			{ $$ = "then"; }
	| ELSE			{ $$ = "else"; }
	| WHILE			{ $$ = "while"; }
	| LOCAL			{ $$ = "local"; }
	| FOR			{ $$ = "for"; }
	| RETURN		{ $$ = "return"; }
	| BREAK			{ $$ = "break"; }
	| VOID			{ $$ = "void"; }
	| CATCH			{ $$ = "catch"; }
	;

opword	: '+'			{ $$ = "+"; }
	| '-'			{ $$ = "-"; }
	| '*'			{ $$ = "*"; }
	| '/'			{ $$ = "/"; }
	| '='			{ $$ = "="; }
	| NEQ			{ $$ = "!="; }
	| LT			{ $$ = "<"; }
	| LE			{ $$ = "<="; }
	| GT			{ $$ = ">"; }
	| GE			{ $$ = ">="; }
	| ANDAND		{ $$ = "&&"; }
	| OROR			{ $$ = "||"; }
	;

%%

/***********************************************************************/
/***********************************************************************/
/***********************************************************************/

#include "list.h"

void init_parser ()
{
    parse_stack = create_stack ();
    init_yylex ();
}


static Node *do_parsing (oneline)
    int oneline;
{
    Node *tree;
    Stack *old_stack = parse_stack;
    
    parse_stack = push_stack (parse_stack);
    tree = NULL;
    do {
	set_prompt1 ();
	reset_yylex ();
	if (yyparse ()) {
		      /* Parsing error.  Assume that the_error has been set. */
	    tree = NULL;
	    break;
	} else {
	    tree = mkappend (nCmd, tree, parse_tree);
	}
    } while (!l_stream->eofmet && !oneline);
    tree = copy_tree (tree);		     /* copy in a permanent location */
    parse_stack = pop_stack (old_stack);
    return tree;
}

/*
 * These function return a tree (maybe NULL) if there was no error,
 * and NULL if there was an error.  They must be called with the_error = 0.
 */



void parse_file_begin (file)
    FILE *file;
{
    l_setinput_file (file);
}

Node *parse_file_next ()
{
    return do_parsing (1); /* one line */
}

void parse_file_end ()
{
    l_popstream ();
}

Node *parse_string (string)
    char *string;
{
    Node *tree;
    l_setinput_string (string);
    tree = do_parsing (0); /* all */
    l_popstream ();
    return tree;
}

void parse_readline_begin ()
{
    l_setinput_readline ();
}

Node *parse_readline_next ()
{
    set_prompt1 ();
    return do_parsing (1); /* one line */
}

void parse_readline_end ()
{
    l_popstream ();
}
