/* readline.c */

#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h> /* for select stuff */

#include "scrio.h"
#include "mem.h"

#include "externs.h"




void init_readline ()
{
}



void (*readline_winch_routine) ();
void (*readline_echo_fun) (/* x, char */);
void (*readline_setcur_fun) (/* x */);
void (*readline_flush_fun) ();
void (*readline_begin_fun) ();
void (*readline_end_fun) ();
int readline_limited;			       /* nonzero means limited area */
int readline_width;

static int pos;			/* virtual pos of cursor in buffer (>= plen) */
static int buflen;
static int begpos;   /* first pos in buffer that is seen on screen (limited) */
static int bufsize;
static char *buf = NULL;


static void echo_string (x, s, sz)
    int x;
    char *s;
    int sz;
{
    int i;
    for (i=0; i<sz; i++)
	(*readline_echo_fun) (x+i, s[i]);
}


static void redraw_buf (force)				     /* when limited */
    int force;
{
    if (buflen < begpos) {
	begpos = buflen;
	force = 1;
    }
    if (pos-begpos < readline_width/4 && begpos > 0) {
	begpos -= readline_width/2;
	if (pos < begpos)
	    begpos = pos - readline_width/4;
	if (begpos < 0)
	    begpos = 0;
	force = 1;
    } else if (pos-begpos > (3*readline_width)/4) {
	begpos += readline_width/2;
	if (pos > begpos + (3*readline_width)/4)
	    begpos = pos - (3*readline_width)/4;
	if (begpos < 0)
	    begpos = 0;
	force = 1;
    }
    if (force) {
	int nbchars = buflen - begpos;
	if (nbchars > readline_width)
	    nbchars = readline_width;
	echo_string (0, buf+begpos, nbchars);
	if (nbchars < readline_width) {
	    int i;
	    for (i=readline_width-nbchars; i>0; i--)
		(*readline_echo_fun) (readline_width-i, ' ');
	}
    }
    (*readline_setcur_fun) (pos-begpos);
    (*readline_flush_fun) ();
}

char *readline (prompt)		  /* returns a string that has to be mfree'd */
    char *prompt;
{
    int plen = strlen (prompt);
    int k;
    static int reachedeof = 0;

    if (reachedeof)
	return NULL;

    (*readline_begin_fun) ();
    if (buf == NULL) {
	buf = mmalloc (bufsize = 100 + plen);
    }
    if (bufsize < plen+5)
	buf = mrealloc (bufsize += plen + 20);
    plen = strlen (prompt);
    memcpy (buf, prompt, plen);
    pos = plen;
    buflen = plen;
    if (!readline_limited) {
	echo_string (0, prompt, plen);
    } else {
	if (plen+1 <= readline_width) {
	    begpos = 0;
	    echo_string (0, prompt, plen);
	} else {
	    begpos = (plen+1) - readline_width;
	    echo_string (0, prompt+begpos, readline_width-1);
	}
	(*readline_setcur_fun) (pos-begpos);
    }
    (*readline_flush_fun) ();
    while (1) {
	k = get_key ();
	if (k == -1) /* eof */
	    break;
	if (k == -2) { /* interrupted */
	    extern int sig_winch_pend;
	    if (readline_limited && sig_winch_pend) {
		(*readline_winch_routine) ();
		redraw_buf (1);
		sig_winch_pend = 0;
	    }
	    continue;
	}
	if (k==10 || k==13) {					   /* return */
	    break;
	}
	if (k==8 || k==127) {				 /* backspace/delete */
	    if (pos>plen) {
		if (pos < buflen)
		    bcopy (buf+pos, buf+pos-1, buflen-pos);
		pos--;
		buflen--;
		if (!readline_limited) {
		    int i;
		    (*readline_echo_fun) (0, 8);
		    for (i=0; i<buflen-pos; i++)
			(*readline_echo_fun) (0, buf[pos+i]);
		    (*readline_echo_fun) (0, ' ');
		    for (i=0; i<buflen-pos+1; i++)
			(*readline_echo_fun) (0, 8);
		    (*readline_flush_fun) ();
		} else {
		    if (pos >= begpos && pos == buflen)
			(*readline_echo_fun) (pos-begpos, ' ');
		    redraw_buf (pos < buflen);
		}
	    }
	    continue;
	}
	if (k=='D'-'@') {
	    if (pos<buflen) {
		if (pos < buflen-1)
		    bcopy (buf+pos+1, buf+pos, buflen-pos-1);
		buflen--;
		if (!readline_limited) {
		    int i;
		    for (i=0; i<buflen-pos; i++)
			(*readline_echo_fun) (0, buf[pos+i]);
		    (*readline_echo_fun) (0, ' ');
		    for (i=0; i<buflen-pos+1; i++)
			(*readline_echo_fun) (0, 8);
		    (*readline_flush_fun) ();
		} else {
		    redraw_buf (1);
		}
	    }
	    continue;
	}
	if (k=='K'-'@') {
	    if (pos<buflen) {
		int old = buflen;
		buflen = pos;
		if (!readline_limited) {
		    int i;
		    for (i=0; i<old-pos; i++)
			(*readline_echo_fun) (0, ' ');
		    for (i=0; i<old-pos; i++)
			(*readline_echo_fun) (0, 8);
		    (*readline_flush_fun) ();
		} else {
		    redraw_buf (1);
		}
	    }
	    continue;
	}
	if (k=='U'-'@') {
	    if (!readline_limited) {
		int i;
		for (i=0; i<buflen-pos; i++)
		    (*readline_echo_fun) (0, ' ');
		for (i=0; i<buflen-pos; i++)
		    (*readline_echo_fun) (0, 8);
		while (pos>plen) {
		    pos--;
		    (*readline_echo_fun) (0, 8);
		    (*readline_echo_fun) (0, ' ');
		    (*readline_echo_fun) (0, 8);
		}
		pos = plen;
		buflen = plen;
		(*readline_flush_fun) ();
	    } else {
		pos = plen;
		buflen = plen;
		redraw_buf (1);
	    }
	    continue;
	}
	if (k == 'B'-'@') {
	    if (pos>plen) {
		pos--;
		if (!readline_limited) {
		    (*readline_echo_fun) (0, 8);
		    (*readline_flush_fun) ();
		} else {
		    redraw_buf (0);
		}
	    }
	    continue;
	}
	if (k == 'A'-'@') {
	    if (!readline_limited) {
		for (; pos>plen; pos--)
		    (*readline_echo_fun) (0, 8);
		(*readline_flush_fun) ();
	    } else {
		pos = plen;
		redraw_buf (0);
	    }
	    continue;
	}
	if (k == 'F'-'@') {
	    if (pos<buflen) {
		pos++;
		if (!readline_limited) {
		    (*readline_echo_fun) (0, buf[pos-1]);
		    (*readline_flush_fun) ();
		} else {
		    redraw_buf (0);
		}
	    }
	    continue;
	}
	if (k == 'E'-'@') {
	    if (!readline_limited) {
		for (; pos<buflen; pos++)
		    (*readline_echo_fun) (0, buf[pos]);
		(*readline_flush_fun) ();
	    } else {
		pos = buflen;
		redraw_buf (0);
	    }
	    continue;
	}
	if (k==9)
	    k = 32;
	if (k>=32 && k<=255) {
	    if (buflen+3 >= bufsize)
		buf = mrealloc (buf, bufsize *= 2);
	    if (pos < buflen)
		bcopy (buf+pos, buf+pos+1, buflen-pos);
	    buf[pos] = k;
	    pos++;
	    buflen++;
	    if (!readline_limited) {
		int i;
		for (i = -1; i<buflen-pos; i++)
		    (*readline_echo_fun) (0, buf[pos+i]);
		for (i=0; i<buflen-pos; i++)
		    (*readline_echo_fun) (0, 8);
		(*readline_flush_fun) ();
	    } else {
		if (pos == buflen)
		    (*readline_echo_fun) (pos-begpos-1, buf[pos-1]);
		redraw_buf (pos < buflen);
	    }
	    continue;
	}
    }

    if (!readline_limited) {
	for (; pos<buflen; pos++)
	    (*readline_echo_fun) (0, buf[pos]);
    }
    (*readline_end_fun) ();
    buf[buflen] = '\0';

    if (k == -1) {
	reachedeof = 1;
	if (pos == 0)
	    return NULL;
    }
    return mstrdup (buf+plen);
}


void add_history ()
{
}




void readline_stdout_echo (x, c)
    int x;
    unsigned char c;
{
    putchar (c);
}

void readline_stdout_setcur ()
{
    fprintf (stderr, "BUG shouldn't call setcut in stdout readline\n");
}

void readline_stdout_flush ()
{
    fflush (stdout);
}

void readline_stdout_begin ()
{
}


void readline_stdout_end ()
{
    putchar ('\n');
    fflush (stdout);
}


