/* screen.c */

#include <stdio.h>

#include "mur.h"		/* global definitions */
#include "scrio.h"
#include "mem.h"
#include "brick.h"
#include "toplevel.h"
#include "list.h"
#include "util.h"
#include "screen.h"


State *state = NULL;
State *previous_state = NULL;

int flag_nocoo = 0;

int scroll_delta_x = 0;
int scroll_delta_y = 0;

List *getvar_pos ()
{
    static char buf1[20];
    static char buf2[20];
    static List list1;
    static List list2;
    sprintf (buf1, "%d", state->pos_x + state->trans_x);
    sprintf (buf2, "%d", state->pos_y + state->trans_y);
    list1.arg = buf1;
    list2.arg = buf2;
    list1.next = &list2;
    list2.next = NULL;
    return &list1;
}

void setvar_pos (l)
    List *l;
{
    long resx, resy;

    if (l == NULL || l->next == NULL || l->next->next != NULL)
	return;
    if (long_from_string (l->arg, &resx)
	|| long_from_string (l->next->arg, &resy))
	return;
    state->trans_x = resx - state->pos_x;
    state->trans_y = resy - state->pos_y;
}


List *getvar_abspos ()
{
    static char buf1[20];
    static char buf2[20];
    static List list1;
    static List list2;
    sprintf (buf1, "%d", state->pos_x);
    sprintf (buf2, "%d", state->pos_y);
    list1.arg = buf1;
    list2.arg = buf2;
    list1.next = &list2;
    list2.next = NULL;
    return &list1;
}

void setvar_abspos ()
{
}


List *getvar_orig ()
{
    static char buf1[20];
    static char buf2[20];
    static List list1;
    static List list2;
    sprintf (buf1, "%d", state->orig_x + state->trans_x);
    sprintf (buf2, "%d", state->orig_y + state->trans_y);
    list1.arg = buf1;
    list2.arg = buf2;
    list1.next = &list2;
    list2.next = NULL;
    return &list1;
}

void setvar_orig (l)
    List *l;
{
    long res;

    if (l == NULL || l->next == NULL || l->next->next != NULL)
	return;
    /* Two separate affectations, too. */
    if (!long_from_string (l->arg, &res))
	state->orig_x = res - state->trans_x;
    if (!long_from_string (l->next->arg, &res))
	state->orig_y = res - state->trans_y;
}


List *getvar_showcoo ()
{
    static List list;
    if (flag_nocoo)
	list.arg = "0";
    else
	list.arg = "1";
    list.next = NULL;
    return &list;
}

void setvar_showcoo (l)
    List *l;
{
    int old = flag_nocoo;
    long val;
    if (l == NULL || l->next != NULL)
	return;
    if (long_from_string (l->arg, &val))
	return;
    flag_nocoo = !val;
    if (flag_nocoo && old == 0 && in_screen) {
	wmove (menuw, 0, cols-29);
	wclrtoeol (menuw);
    }
}

List *getvar_size ()
{
    static char buf1[20];
    static char buf2[20];
    static List list1;
    static List list2;
    sprintf (buf1, "%d", cols);
    sprintf (buf2, "%d", lines);
    list1.arg = buf1;
    list2.arg = buf2;
    list1.next = &list2;
    list2.next = NULL;
    return &list1;
}

void setvar_size ()
{
}


List *getvar_scrolldelta ()
{
    static char buf1[20];
    static char buf2[20];
    static List list1;
    static List list2;
    sprintf (buf1, "%d", scroll_delta_x);
    sprintf (buf2, "%d", scroll_delta_y);
    list1.arg = buf1;
    list2.arg = buf2;
    list1.next = &list2;
    list2.next = NULL;
    return &list1;
}

void setvar_scrolldelta (l)
    List *l;
{
    long res;
    if (l == NULL || l->next == NULL || l->next->next != NULL)
	return;
    if (!long_from_string (l->arg, &res) && res >= 0)
	scroll_delta_x = res;
    if (!long_from_string (l->next->arg, &res) && res >= 0)
	scroll_delta_y = res;
}









static void draw_region (sx, sy, w, h, flag)
    long sx, sy, w, h;
    int flag;					    /* 0=only touched, 1=all */
{
    long x, y;
    long bx, by;
    long dx, dy;
    long nbx, nby;
    Brick *br;
	
    x = sx + state->orig_x;
    y = sy + state->orig_y;
    bx = Abs_brick_from_x (x);
    by = Abs_brick_from_y (y);
    dx = x & X_MASK;
    dy = y & Y_MASK;
    nbx = Abs_brick_x (((x+w-1) >> X_SHIFT) - bx + 1);
    nby = Abs_brick_y (((y+h-1) >> Y_SHIFT) - by + 1);

    if (flag == 0) {					     /* Only touched */
	for (br = bricks_touched; br != NULL; br = br->next_touched) {
	    long brx = br->x;
	    long bry = br->y;
	    if (IsInside (brx, bry, bx, by, nbx, nby)) {
		long posx = (brx << X_SHIFT) - x;     /* pos brick in region */
		long posy = (bry << Y_SHIFT) - y;
		long begx = (posx >= 0) ? 0 : -posx;	  /* offset in brick */
		long begy = (posy >= 0) ? 0 : -posy;
		long endx = (posx+X_BRICK) >= w ? w-posx : X_BRICK; /* end+1 */
		long endy = (posy+Y_BRICK) >= h ? h-posy : Y_BRICK;
		long width = endx - begx;
		long scrx = sx + posx + begx;
		long scry = sy + posy;
		char line [X_BRICK+1];
		long j;
		for (j=begy; j<endy; j++) {
		    if (br->touched_line [j]) {
			memcpy (line, br->line [j] + begx, width);
			line [width] = '\0';
			mvwaddstr (mainw, scry+j, scrx, line);
		    }
		}
	    }
	}
    } else /* if (flag == 1) */ {				      /* All */
	long width, height, first_width, last_width, last_height;
	long posx, posy;
	long j, i, jj;

	add_burst_zone (bx, by, nbx, nby);
	read_brick_queue ();
	last_height = (nby == 1) ? h : (((y+h-1) & Y_MASK) + 1);
	first_width = X_BRICK-dx;
	last_width = (nbx == 1) ? w : (((x+w-1) & X_MASK) + 1);
	posy = dy;
	height = Y_BRICK-dy;
	for (j=0; j<nby; j++) {
	    long scry = sy+j*Y_BRICK-dy+posy;
	    if (j == nby-1) height = last_height;
	    posx = dx;
	    width = first_width;
	    for (i=0; i<nbx; i++) {
		long scrx = sx+i*X_BRICK-dx+posx;
		if (i == nbx-1) width = last_width;
		br = get_brick (bx+i, by+j);
		for (jj=0; jj<height; jj++) {
		    char line [X_BRICK+1];
		    memcpy (line, br->line [posy + jj] + posx, width);
		    line [width] = '\0';
		    mvwaddstr (mainw, scry+jj, scrx, line);
		}
		posx = 0;
		width = X_BRICK;
	    }
	    posy = 0;
	    height = Y_BRICK;
	}
    }
}




static void del_lines_up (n)
    long n;
{
    long j;

    wmove (mainw, 0, 0);
    for (j=0; j<n; j++)
	wdeleteln (mainw);
}


static void del_lines_dn (n)
    long n;
{
    long j;

    for (j=0; j<n; j++) {
	wmove (mainw, j, 0);
	winsertln (mainw);
    }
}


static void del_lines_lt (n)
    long n;
{
    long i, j;

    for (j=0; j<lines; j++) {
	wmove (mainw, j, 0);
	for (i=0; i<n; i++)
	    wdelch (mainw);
    }
}


static void del_lines_rt (n)
    long n;
{
    long i, j;

    for (j=0; j<lines; j++) {
	wmove (mainw, j, 0);
	for (i=0; i<n; i++)
	    winsch (mainw, ' ');
    }
}


void print_coo (flag)				    /* 1=force if reasonable */
    int flag;
{
    if (   (state->pos_x + state->trans_x
	    != previous_state->pos_x + previous_state->trans_x)
	|| (state->pos_y + state->trans_y
	    != previous_state->pos_y + previous_state->trans_y)
	|| state->newcoo
	|| flag) {
	if (!flag_nocoo && mode == MODE_INTERACT) {
	    char tmp[50];
	    sprintf (tmp, " x=%d y=%d",
		     state->pos_x + state->trans_x,
		     state->pos_y + state->trans_y);
	    mvwaddstr (menuw, 0, cols-29, tmp);
	    wclrtoeol (menuw);
	    state->newcoo = 0;
	} else {
	    state->newcoo = 1;
	}
    }
}

void position_cursor ()
{
    wmove (mainw, state->pos_y - state->orig_y, state->pos_x - state->orig_x);
}



void reprint_screen (flag)
    int flag;						     /* 0=new, 1=all */
{
    long dx, dy;
    long x, y, w, h;
    Brick *br;

    if (newminilines != minishelllines) {
	if (newminilines > LINES-1)
	    newminilines = LINES-1;
	if (newminilines < 1)
	    newminilines = 1;
	if (newminilines != minishelllines) {
	    int oldminilines = minishelllines;
	    flag = 1;
	    delwin (mainw);
	    delwin (menuw);
	    minishelllines = newminilines;
	    get_scr_size ();
	    newminilines = minishelllines;
	    mainw = subwin (stdscr, lines, cols, 0, 0);
	    if (mainw == (WINDOW *) ERR)
		mainw = subwin (stdscr, LINES, COLS, 0, 0);  /* if too small */
	    menuw = subwin (stdscr, minishelllines, cols, lines, 0);
	    if (menuw == (WINDOW *) ERR)
		menuw = subwin (stdscr, LINES, COLS, 0, 0);  /* if too small */
	    leaveok (mainw, FALSE);
	    leaveok (menuw, FALSE);
	    if (newminilines > oldminilines) {
		int i; 
		for (i = newminilines - oldminilines; i; i--) {
		    wmove (menuw, i-1, 0);

		    wclrtoeol (menuw);
		}
		if (!flag_nocoo) {
		    wmove (menuw, newminilines-oldminilines, cols-29);
		    wclrtoeol (menuw);
		}
	    }
	    state->newcoo = 1;
	}
    }

    dx = state->pos_x - state->orig_x;
    if (dx >= (cols*3)/2)
	state->orig_x = state->pos_x - cols/2;
    else if (dx >= cols) {
	state->orig_x = state->pos_x - cols + 1 + scroll_delta_x;
	if (state->pos_x - state->orig_x < 0)
	    state->orig_x = state->pos_x;
    } else if (dx < -cols/2)
	state->orig_x = state->pos_x - cols/2;
    else if (dx < 0) {
	state->orig_x = state->pos_x - scroll_delta_x;
	if (state->pos_x - state->orig_x >= cols)
	    state->orig_x = state->pos_x - cols + 1;
    }

    dy = state->pos_y - state->orig_y;
    if (dy >= (lines*3)/2)
	state->orig_y = state->pos_y - lines/2;
    else if (dy >= lines) {
	state->orig_y = state->pos_y - lines + 1 + scroll_delta_y;
	if (state->pos_y - state->orig_y < 0)
	    state->orig_y = state->pos_y;
    } else if (dy < -lines/2)
	state->orig_y = state->pos_y - lines/2;
    else if (dy < 0) {
	state->orig_y = state->pos_y - scroll_delta_y;
	if (state->pos_y - state->orig_y >= lines)
	    state->orig_y = state->pos_y - lines + 1;
    }

    dx = state->orig_x - previous_state->orig_x;
    dy = state->orig_y - previous_state->orig_y;

    if (flag == 1 || state->warped
	|| dx > cols/2 || dx < -cols/2
	|| dy > lines/2 || dy < -lines/2) {
	draw_region (0, 0, cols, lines, 1);
	w = 0;
    } else if (dx >= 0 && dy >= 0) {
	if (dy) del_lines_up (dy);
	if (dx) del_lines_lt (dx);
	if (dx) draw_region (cols-dx, 0, dx, lines-dy, 1);
	if (dy) draw_region (0, lines-dy, cols, dy, 1);
	x = 0;
	y = 0;
	w = cols-dx;
	h = lines-dy;
    } else if (dx >= 0 && dy < 0) {
	del_lines_dn (-dy);
	if (dx) del_lines_lt (dx);
	draw_region (0, 0, cols, -dy, 1);
	if (dx) draw_region (cols-dx, -dy, dx, lines+dy, 1);
	x = 0;
	y = -dy;
	w = cols-dx;
	h = lines+dy;
    } else if (dx < 0 && dy >= 0) {
	if (dy) del_lines_up (dy);
	del_lines_rt (-dx);
	draw_region (0, 0, -dx, lines-dy, 1);
	if (dy) draw_region (0, lines-dy, cols, dy, 1);
	x = -dx;
	y = 0;
	w = cols+dx;
	h = lines-dy;
    } else /* if (dx < 0 && dy < 0) */ {
	del_lines_dn (-dy);
	del_lines_rt (-dx);
	draw_region (0, 0, cols, -dy, 1);
	draw_region (0, -dy, -dx, lines+dy, 1);
	x = -dx;
	y = -dy;
	w = cols+dx;
	h = lines+dy;
    }

    if (w && h) {
	draw_region (x, y, w, h, 0); /* only touched */
    }

    /*
     * Now clear touched flags.
     */
    for (br = bricks_touched; br != NULL; br = br->next_touched) {
	long j;
	for (j=0; j<Y_BRICK; j++)
	    br->touched_line [j] = 0;
	br->touched = 0;
    }
    bricks_touched = NULL;

    print_coo (flag);

    /*
     * Change previous state.
     */
    previous_state->pos_x = state->pos_x;
    previous_state->pos_y = state->pos_y;
    previous_state->trans_x = state->trans_x;
    previous_state->trans_y = state->trans_y;
    previous_state->orig_x = state->orig_x;
    previous_state->orig_y = state->orig_y;
    state->warped = 0;

    position_cursor ();
}


void menu_scroll ()
{
    wmove (menuw, 0, 0);
    wdeleteln (menuw);
    wmove (menuw, minishelllines-1, 0);
    state->newcoo = 1;
}

