/* toplevel.c */

#include <stdio.h>
#include <stdlib.h>

#include "mur.h"
#include "scrio.h"
#include "brick.h"
#include "commands.h"
#include "keymap.h"
#include "funs.h"
#include "mem.h"
#include "screen.h"
#include "error.h"
#include "parse.h"
#include "list.h"
#include "readline.h"
#include "util.h"
#include "toplevel.h"

#include "externs.h"



static int modeline_full = 0;


enum toplevel_mode mode = MODE_SHELL;
enum toplevel_mode newmode = MODE_SHELL;

void do_sig_winch ();


/************************************************************************/



void find_orig_trans ()
{
    int miny = Y_BRICK;
    int maxy;	
    int empty = 1;
    int i, j;
    Brick *br;
    br = current_brick = get_brick (0, 0);
    for (j = 0; j<Y_BRICK; j++) {
	for (i = 0; i<X_BRICK; i++) {
	    if (br->line[j][i] != ' ') {
		if (j < miny)
		    miny = j;
		maxy = j;
		empty = 0;
		break;
	    }
	}
    }
    if (empty) { /* shouldn't happen */
	state->pos_x = X_BRICK/2;
	state->pos_y = Y_BRICK/2;
    } else {
	extern double drand48 ();
	int jj = miny+(int)((maxy-miny+1) * drand48 ());
	int ii;
	int minx = X_BRICK;
	int maxx;
	int found = 0;
	while (1) {
	    for (i = 0; i<X_BRICK; i++) {
		if (br->line[jj][i] != ' ') {
		    found = 1;
		    if (i < minx)
			minx = i;
		    maxx = i;
		}
	    }
	    if (found)
		break;
	    jj++;
	}
	ii = minx+(int)((maxx-minx+1) * drand48 ());
	while (1) {
	    if (br->line[jj][ii] != ' ')
		break;
	    ii++;
	}
	state->pos_x = ii;
	state->pos_y = jj;
    }
    state->trans_x = - state->pos_x;
    state->trans_y = - state->pos_y;
}










void init_state ()
{
    if (previous_state == NULL)
	previous_state = New (State);
    previous_state->orig_x = 0; /* useless */
    previous_state->orig_y = 0;
    previous_state->mapping_mode = 0;

    if (state == NULL)
	state = New (State);
    find_orig_trans ();
    state->orig_x = state->pos_x-cols/2;
    state->orig_y = state->pos_y-lines/2;
    state->lastmove_x = state->pos_x;
    state->lastmove_y = state->pos_y;
    state->mapping_mode = 0;
    state->newcoo = 1;
    state->warped = 1;
}


/************************************************************************/


List *getvar_mode ()
{
    static List list;
    switch (newmode) {
    case MODE_SHELL:
	list.arg = "shell";
	break;
    case MODE_INTERACT:
	list.arg = "interactive";
	break;
    case MODE_MINISHELL:
	list.arg = "minishell";
	break;
    default:
	list.arg = "";
	break;
    }
    list.next = NULL;
    return &list;
}

void setvar_mode (l)
    List *l;
{
    if (l == NULL || l->next != NULL)
	return;
    if (STREQ ("shell", l->arg))
	newmode = MODE_SHELL;
    else if (STREQ ("interactive", l->arg))
	newmode = MODE_INTERACT;
    else if (STREQ ("minishell", l->arg))
	newmode = MODE_MINISHELL;
}

List *getvar_minilines ()
{
    static char buf[20];
    static List list;
    sprintf (buf, "%d", newminilines);
    list.arg = buf;
    list.next = NULL;
    return &list;
}

void setvar_minilines (l)
    List *l;
{
    if (l == NULL || l->next != NULL)
	return;
    newminilines = atoi (l->arg);
    if (newminilines < 1)
	newminilines = 1;
}

List *getvar_mousestate ()
{
    static List list;
    if (want_mouse)
	list.arg = "1";
    else
	list.arg = "0";
    list.next = NULL;
    return &list;
}

void setvar_mousestate (l)
    List *l;
{
    long val;
    if (l == NULL || l->next != NULL)
	return;
    if (long_from_string (l->arg, &val))
	return;
    if (!is_xterm)
	return;
    switch (val) {
    case 0:
	want_mouse = 0;
	if (mode == MODE_INTERACT)
	    set_mouse_off ();
	break;
    case 1:
	want_mouse = 1;
	if (mode == MODE_INTERACT)
	    set_mouse_on ();
	break;
    }
}

List *getvar_slowmouse ()
{
    static List list;
    if (slow_mouse)
	list.arg = "1";
    else
	list.arg = "0";
    list.next = NULL;
    return &list;
}

void setvar_slowmouse (l)
    List *l;
{
    long val;
    if (l == NULL || l->next != NULL)
	return;
    if (long_from_string (l->arg, &val))
	return;
    slow_mouse = (val != 0);
}










static void interact_sig_winch ()
{
    stty_running ();
    exit_scr ();
    init_scr ();
    stty_getchar ();
}


static int get_key_interact ()
{
    int key;
    sig_winch_routine = interact_sig_winch;
    do {
	sig_winch_ok = 1;
	key = get_key ();
	sig_winch_ok = 0;
    } while (key == -2);
    return key;
}


/* limited readline stuff */

int limited_readline_x;
int limited_readline_y;

void limited_readline_winch_routine ()
{
    stty_running ();
    exit_scr ();
    init_scr ();
    stty_getchar ();
    readline_width = cols;
    limited_readline_y = minishelllines - 1;
}

void limited_readline_echo (x, c)
    int x;
    unsigned char c;
{
    mvwaddch (menuw, limited_readline_y, limited_readline_x + x, c);
}

void limited_readline_setcur (x)
    int x;
{
    wmove (menuw, limited_readline_y, limited_readline_x + x);
}

void limited_readline_flush ()
{
    wrefresh (menuw);
}

void limited_readline_begin ()
{
    show_message ("", 2); /* clear last line */
}

void limited_readline_end ()
{
    wmove (menuw, minishelllines-1, 0);
    show_message ("", 1); /* add a pending newline */
}




void show_message (s, newline)
    char *s;
    int newline;
{
    static int curx = 0;
    static int pending_newline = 0;
    switch (mode) {
    case MODE_INTERACT:
    case MODE_MINISHELL: {
	char *line;
	int sz;

	if (newline == 1) {
	    pending_newline = 1;
	} else if (newline == 2) { /* clear last line */
	    if (curx != 0 || pending_newline) {
		menu_scroll ();
		curx = 0;
	    }
	    pending_newline = 0;
	}
	sz = strlen (s);
	line = mmalloc (cols+1);
	while (sz) {
	    int i;
	    int ocurx;
	    if (pending_newline) {
		menu_scroll ();
		pending_newline = 0;
		curx = 0;
	    }
	    ocurx = curx;
	    for (i=0; i<sz && curx < cols; i++) {
		char c = s[i];
		if (c == '\n')
		    break;
		if (c == 9) {
		    int n = ((curx+8) & ~7) - curx;
		    for (; n>=0 && curx < cols; n--)
			line[curx++] = ' ';
		    continue;
		}
		if (c < ' ')
		    c = ' ';
		line[curx++] = c;
	    }
	    line[curx] = '\0';
	    mvwaddstr (menuw, minishelllines-1, ocurx, line+ocurx);
	    if (curx == cols)
		pending_newline = 1;
	    if (s[i] == '\n') {
		pending_newline = 1;
		i++;
	    }
	    s += i;
	    sz -= i;
	}
	mfree (line);
	break;
    }
    case MODE_SHELL:
	if (newline == 1)
	    printf ("\n%s", s);
	else 
	    printf ("%s", s);
	fflush (stdout);
	curx = 0;
	pending_newline = 2; /* for when we come back to interact */
	break;
    default: /* just to make gcc -Wall happy */
	;
    }
}




static void interact ()
{
    print_coo (1);
    wrefresh (menuw);
    position_cursor ();
    wrefresh (mainw);
    modeline_full = 0;
    while (1) {
	Keymap *keymap = the_keymap;
	int olduse;
	Stack *old_stack;
	Keycommand *kc;
	int c;

	while (1) {
	    Keymap *newkeymap;
	    int res;
	    c = get_key_interact ();
	    if (c == -1) /* eof */
		break;
	    res = walk_keymap (keymap, c, &newkeymap, &kc);
	    if (res == 1) {					   /* keymap */
		keymap = newkeymap;
	    } else if (res == 2) {			       /* keycommand */
		break;
	    } else {
		beep ();
		keymap = the_keymap;
	    }
	}
	if (c != -1) {
	    old_stack = command_stack;
	    command_stack = push_stack (command_stack);
	    olduse = kc->inuse;
	    kc->inuse = 1;
	    exec_tree (kc->tree);
	    kc->inuse = olduse;
	    command_stack = pop_stack (old_stack);
	    free_pending_functions ();
	    free_pending_kcs (); /* in keymaps */
	}
	if (the_error) {
	    if (the_error == ERR_TOPLEVEL) {
		reprint_screen (0);
		break;
	    }
	    show_message (the_error_message, 1);
	    show_message ("", 1);
	    wrefresh (menuw);
	    modeline_full = 1;
	    no_error ();
	} else {
	    if (modeline_full) {
		modeline_full = 0;
		menu_scroll ();
		wrefresh (menuw);
	    }
	}
	if (!key_pressed ()) {
	    reprint_screen (0);
	    wrefresh (menuw); /* for messages */
	    wrefresh (mainw); /* in second, for cursor */
	}
    }
}

static void shell ()
{
    parse_readline_begin ();
    while (1) {
	Stack *old_stack;
	Node *tree;

	readline_limited = 0;
	readline_echo_fun = readline_stdout_echo;
	readline_setcur_fun = readline_stdout_setcur;	 /* in case I bugged */
	readline_flush_fun = readline_stdout_flush;
	readline_begin_fun = readline_stdout_begin;
	readline_end_fun = readline_stdout_end;
	tree = parse_readline_next ();
	if (!the_error) {
	    old_stack = command_stack;
	    command_stack = push_stack (command_stack);
	    exec_tree (tree);
	    command_stack = pop_stack (old_stack);
	    free_tree (tree);
	    free_pending_functions ();
	}
	if (the_error) {
	    if (the_error == ERR_TOPLEVEL)
		break;
	    show_message (the_error_message, 0);
	    show_message ("", 1);
	    no_error ();
	}
    }
    parse_readline_end ();
}


static void minishell ()
{
    parse_readline_begin ();
    while (1) {
	Node *tree;
	Stack *old_stack;

	readline_limited = 1;
	readline_width = cols;
	limited_readline_x = 0;
	readline_winch_routine = limited_readline_winch_routine;
	readline_echo_fun = limited_readline_echo;
	readline_setcur_fun = limited_readline_setcur;
	readline_flush_fun = limited_readline_flush;
	readline_begin_fun = limited_readline_begin;
	readline_end_fun = limited_readline_end;
	limited_readline_y = minishelllines - 1;
	tree = parse_readline_next ();
	if (!the_error) {
	    old_stack = command_stack;
	    command_stack = push_stack (command_stack);
	    exec_tree (tree);
	    command_stack = pop_stack (old_stack);
	    free_tree (tree);
	    free_pending_functions ();
	}
	reprint_screen (0);
	if (the_error) {
	    if (the_error == ERR_TOPLEVEL)
		break;
	    show_message (the_error_message, 1);
	    show_message ("", 1);
	    no_error ();
	}
	wrefresh (menuw);
	wrefresh (mainw);
    }
    parse_readline_end ();
}



void toplevel ()
{
    while (1) {
	no_error ();
	if (newmode != mode) {
	    if ((newmode == MODE_INTERACT || newmode == MODE_MINISHELL)
		&& mode == MODE_SHELL) {
		init_scr ();
		reprint_screen (1);
		wrefresh (mainw);
	    }
	    if (newmode == MODE_INTERACT && want_mouse)
		set_mouse_on ();
	    if (newmode != MODE_INTERACT)
		set_mouse_off ();
	    if ((newmode == MODE_SHELL || newmode == MODE_EXIT)
		&& (mode == MODE_INTERACT || mode == MODE_MINISHELL)) {
		show_message ("", 2);
		exit_scr ();
	    }
	    mode = newmode;
	}
	if (mode == MODE_INTERACT)
	    interact ();
	else if (mode == MODE_SHELL)
	    shell ();
	else if (mode == MODE_MINISHELL)
	    minishell ();
	else if (mode == MODE_EXIT)
	    return;
    }
}


