/* tree.c */

#include <stdio.h>

#include "tree.h"
#include "mem.h"

#include "externs.h"


extern Stack *parse_stack;

/*
 * freelist for Node
 */

static Node *Node_freelist = NULL;


static Node *new_Node ()
{
    Node *n;
    if (Node_freelist) {
	n = Node_freelist;
	Node_freelist = n->u[0].p;
    } else
	n = New (Node);
    return n;
}

static void free_Node (n)
    Node *n;
{
    n->u[0].p = Node_freelist;
    Node_freelist = n;
}





Node *mknode (t)					     /* in the stack */
    Nodetype t;
{
    Node *n;
    n = SNew (parse_stack, Node);
    n->type = t;
    return n;
}


Node *mkappend (type, n1, n2)
    Nodetype type;
    Node *n1, *n2;
{
    if (n1 != NULL) {
	if (n2 != NULL) {
	    Node *n = mknode (type);
	    n->u[0].p = n1;
	    n->u[1].p = n2;
	    return n;
	} else
	    return n1;
    } else
	return n2;
}


Node *copy_tree (n)			   /* copy tree in a permanent place */
    Node *n;
{
    Node *r;

    if (n == NULL) {
	return NULL;
    }
    r = new_Node ();			/* NOT in stack, we want to keep it. */
    r->type = n->type;
    switch (n->type) {
    case nCmd:
    case nVarsub:
    case nConcat:
    case nArgs:
    case nAdd:
    case nSub:
    case nMul:
    case nDiv:
    case nEq:
    case nNeq:
    case nLt:
    case nLe:
    case nGt:
    case nGe:
    case nAndand:
    case nOror:
    case nFn:
    case nMap:
    case nSetvar:
    case nWhile:
    case nBackq:
    case nLocal:
	r->u[0].p = copy_tree (n->u[0].p);
	r->u[1].p = copy_tree (n->u[1].p);
	return r;
    case nIf:
    case nFor:
	r->u[0].p = copy_tree (n->u[0].p);
	r->u[1].p = copy_tree (n->u[1].p);
	r->u[2].p = copy_tree (n->u[2].p);
	return r;
    case nWord:
	r->u[0].s = mstrdup (n->u[0].s);
	return r;
    case nVar:
    case nCount:
    case nFlat:
    case nBrace:
    case nReturn:
    case nBreak:
    case nArith:
    case nVoid:
    case nCatch:
	r->u[0].p = copy_tree (n->u[0].p);
	return r;
    default:
	fprintf (stderr, "forgotten case %d in copy_tree\n", n->type);
	exit (2);
    }
    /*NOTREACHED*/
}



void free_tree (n)	       /* free a tree that was permanently allocated */
    Node *n;
{
    if (n == NULL)
	return;
    switch (n->type) {
    case nCmd:
    case nVarsub:
    case nConcat:
    case nArgs:
    case nAdd:
    case nSub:
    case nMul:
    case nDiv:
    case nEq:
    case nNeq:
    case nLt:
    case nLe:
    case nGt:
    case nGe:
    case nAndand:
    case nOror:
    case nFn:
    case nMap:
    case nSetvar:
    case nWhile:
    case nBackq:
    case nLocal:
	free_tree (n->u[0].p);
	free_tree (n->u[1].p);
	free_Node (n);
	return;
    case nIf:
    case nFor:
	free_tree (n->u[0].p);
	free_tree (n->u[1].p);
	free_tree (n->u[2].p);
	free_Node (n);
	return;
    case nWord:
	mfree (n->u[0].s);
	free_Node (n);
	return;
    case nVar:
    case nCount:
    case nFlat:
    case nBrace:
    case nReturn:
    case nBreak:
    case nArith:
    case nVoid:
    case nCatch:
	free_tree (n->u[0].p);
	free_Node (n);
	return;
    default:
	fprintf (stderr, "forgotten case %d in copy_tree\n", n->type);
	exit (2);
    }
    /*NOTREACHED*/
}



void print_tree (n)
    Node *n;
{
    if (n == NULL) {
	printf ("NULL");
	return;
    }
    switch (n->type) {
    case nCmd:
	printf ("(cmd ");
	print_tree (n->u[0].p);
	printf (" ");
	print_tree (n->u[1].p);
	printf (")");
	break;
    case nWord:
	printf ("(word %s)", n->u[0].s);
	break;
    case nVar:
	printf ("(var ");
	print_tree (n->u[0].p);
	printf (")");
	break;
    case nVarsub:
	printf ("(varsub ");
	print_tree (n->u[0].p);
	printf (" ");
	print_tree (n->u[1].p);
	printf (")");
	break;
    case nWhile:
	printf ("(while ");
	print_tree (n->u[0].p);
	printf (" ");
	print_tree (n->u[1].p);
	printf (")");
	break;
    case nIf:
	printf ("(if ");
	print_tree (n->u[0].p);
	printf (" ");
	print_tree (n->u[1].p);
	printf (" ");
	print_tree (n->u[2].p);
	printf (")");
	break;
    case nFor:
	printf ("(for ");
	print_tree (n->u[0].p);
	printf (" ");
	print_tree (n->u[1].p);
	printf (" ");
	print_tree (n->u[2].p);
	printf (")");
	break;
    case nCount:
	printf ("(count ");
	print_tree (n->u[0].p);
	printf (")");
	break;
    case nFlat:
	printf ("(flat ");
	print_tree (n->u[0].p);
	printf (")");
	break;
    case nBackq:
	printf ("(backq ");
	print_tree (n->u[0].p);
	printf (" ");
	print_tree (n->u[1].p);
	printf (")");
	break;
    case nLocal:
	printf ("(local ");
	print_tree (n->u[0].p);
	printf (" ");
	print_tree (n->u[1].p);
	printf (")");
	break;
    case nBrace:
	printf ("(brace ");
	print_tree (n->u[0].p);
	printf (")");
	break;
    case nReturn:
	printf ("(return ");
	print_tree (n->u[0].p);
	printf (")");
	break;
    case nBreak:
	printf ("(break ");
	print_tree (n->u[0].p);
	printf (")");
	break;
    case nConcat:
	printf ("(concat ");
	print_tree (n->u[0].p);
	printf (" ");
	print_tree (n->u[1].p);
	printf (")");
	break;
    case nArgs:
	printf ("(args ");
	print_tree (n->u[0].p);
	printf (" ");
	print_tree (n->u[1].p);
	printf (")");
	break;
    case nFn:
	printf ("(fn ");
	print_tree (n->u[0].p);
	printf (" ");
	print_tree (n->u[1].p);
	printf (")");
	break;
    case nMap:
	printf ("(map ");
	print_tree (n->u[0].p);
	printf (" ");
	print_tree (n->u[1].p);
	printf (")");
	break;
    case nSetvar:
	printf ("(setvar ");
	print_tree (n->u[0].p);
	printf (" ");
	print_tree (n->u[1].p);
	printf (")");
	break;
    case nArith:
	printf ("(arith ");
	print_tree (n->u[0].p);
	printf (")");
	break;
    case nVoid:
	printf ("(void ");
	print_tree (n->u[0].p);
	printf (")");
	break;
    case nCatch:
	printf ("(catch ");
	print_tree (n->u[0].p);
	printf (")");
	break;
    case nAdd:
	printf ("(add ");
	print_tree (n->u[0].p);
	printf (" ");
	print_tree (n->u[1].p);
	printf (")");
	break;
    case nSub:
	printf ("(sub ");
	print_tree (n->u[0].p);
	printf (" ");
	print_tree (n->u[1].p);
	printf (")");
	break;
    case nMul:
	printf ("(mul ");
	print_tree (n->u[0].p);
	printf (" ");
	print_tree (n->u[1].p);
	printf (")");
	break;
    case nDiv:
	printf ("(div ");
	print_tree (n->u[0].p);
	printf (" ");
	print_tree (n->u[1].p);
	printf (")");
	break;
    case nEq:
	printf ("(eq ");
	print_tree (n->u[0].p);
	printf (" ");
	print_tree (n->u[1].p);
	printf (")");
	break;
    case nNeq:
	printf ("(neq ");
	print_tree (n->u[0].p);
	printf (" ");
	print_tree (n->u[1].p);
	printf (")");
	break;
    case nLt:
	printf ("(lt ");
	print_tree (n->u[0].p);
	printf (" ");
	print_tree (n->u[1].p);
	printf (")");
	break;
    case nLe:
	printf ("(le ");
	print_tree (n->u[0].p);
	printf (" ");
	print_tree (n->u[1].p);
	printf (")");
	break;
    case nGt:
	printf ("(gt ");
	print_tree (n->u[0].p);
	printf (" ");
	print_tree (n->u[1].p);
	printf (")");
	break;
    case nGe:
	printf ("(ge ");
	print_tree (n->u[0].p);
	printf (" ");
	print_tree (n->u[1].p);
	printf (")");
	break;
    case nAndand:
	printf ("(andand ");
	print_tree (n->u[0].p);
	printf (" ");
	print_tree (n->u[1].p);
	printf (")");
	break;
    case nOror:
	printf ("(oror ");
	print_tree (n->u[0].p);
	printf (" ");
	print_tree (n->u[1].p);
	printf (")");
	break;
    default:
	fprintf (stderr, "forgotten case %d in copy_tree\n", n->type);
	exit (2);
    }
}

