/* variable.c */


#include "mem.h"
#include "hash.h"
#include "list.h"
#include "util.h"
#include "error.h"
#include "variable.h"




extern void setvar_mode ();
extern List *getvar_mode ();
extern void setvar_minilines ();
extern List *getvar_minilines ();
extern void setvar_mousepos ();
extern List *getvar_mousepos ();
extern void setvar_mousestate ();
extern List *getvar_mousestate ();
extern void setvar_slowmouse ();
extern List *getvar_slowmouse ();
extern void setvar_pos ();
extern List *getvar_pos ();
extern void setvar_abspos ();
extern List *getvar_abspos ();
extern void setvar_orig ();
extern List *getvar_orig ();
extern void setvar_size ();
extern List *getvar_size ();
extern void setvar_showcoo ();
extern List *getvar_showcoo ();
extern void setvar_scrolldelta ();
extern List *getvar_scrolldelta ();



static Hashtable *variable_table;


/*
 * freelist for Variable
 */

static Variable *Variable_freelist = NULL;

static Variable *new_Variable ()
{
    Variable *v;
    if (Variable_freelist) {
	v = Variable_freelist;
	Variable_freelist = v->next;
    } else
	v = New (Variable);
    return v;
}

static void free_Variable (v)
    Variable *v;
{
    v->next = Variable_freelist;
    Variable_freelist = v;
}








static List *get_variable (name)	      /* doesn't allocate the result */
    char *name;
{
    Bucket *bucket;

    bucket = search_hashtable (variable_table, name);
    if (bucket != NULL) {
	Variable *var = (Variable *) (bucket->data);
	return var->list;
    } else
	return NULL;
}


List *fetch_variable (name)			/* treats special cases : $1 */
    char *name;
{
    int c = *name;
    long i;

    if (c >= '0' && c <= '9' && (i = ulong_from_string (name)) >= 0) {
	List *l = get_variable ("*");
	List *ll;
	while (i && l != NULL) {
	    i--;
	    l = l->next;
	}
	if (l == NULL)
	    return NULL;
	ll = snew_List ();
	ll->next = NULL;
	ll->arg = l->arg;
	return ll;
    }
    if (STREQ ("_mode", name)) return (getvar_mode ());
    if (STREQ ("_minilines", name)) return (getvar_minilines ());
    if (STREQ ("_mousepos", name)) return (getvar_mousepos ());
    if (STREQ ("_mousestate", name)) return (getvar_mousestate ());
    if (STREQ ("_slowmouse", name)) return (getvar_slowmouse ());
    if (STREQ ("_pos", name)) return (getvar_pos ());
    if (STREQ ("_abspos", name)) return (getvar_abspos ());
    if (STREQ ("_orig", name)) return (getvar_orig ());
    if (STREQ ("_size", name)) return (getvar_size ());
    if (STREQ ("_showcoo", name)) return (getvar_showcoo ());
    if (STREQ ("_scrolldelta", name)) return (getvar_scrolldelta ());
    return get_variable (name);
}




void set_variable (name, list, push)			      /* copies list */
    char *name;
    List *list;
    int push;
{
    Bucket *bucket;

    list = permanent_copy_of_list (list);
    bucket = search_hashtable (variable_table, name);
    if (bucket != NULL) {
	Variable *var = (Variable *) (bucket->data);
	if (push) {
	    Variable *newvar = new_Variable ();
	    newvar->next = var;
	    newvar->list = list;
	    bucket->data = (char *) newvar;
	} else {
	    free_permanent_list (var->list);
	    var->list = list;
	}
    } else {
	Variable *newvar = new_Variable ();
	newvar->next = NULL;
	newvar->list = list;
	insert_hashtable (variable_table, name, (char *) newvar);
    }
    return;
}


void store_variable (name, list, push)
    char *name;
    List *list;
    int push;
{
    int c = *name;

    if (c >= '0' && c <= '9' && ulong_from_string (name) >= 0) {
	set_error (ERR_EXEC, "can't affect numeric variables");
	return;
    }
	 if (STREQ ("_mode", name)) setvar_mode (list);
    else if (STREQ ("_minilines", name)) setvar_minilines (list);
    else if (STREQ ("_mousepos", name)) setvar_mousepos (list);
    else if (STREQ ("_mousestate", name)) setvar_mousestate (list);
    else if (STREQ ("_slowmouse", name)) setvar_slowmouse (list);
    else if (STREQ ("_pos", name)) setvar_pos (list);
    else if (STREQ ("_abspos", name)) setvar_abspos (list);
    else if (STREQ ("_orig", name)) setvar_orig (list);
    else if (STREQ ("_size", name)) setvar_size (list);
    else if (STREQ ("_showcoo", name)) setvar_showcoo (list);
    else if (STREQ ("_scrolldelta", name)) setvar_scrolldelta (list);
    else set_variable (name, list, push);
}



void delete_variable (name)
    char *name;
{
    Bucket *bucket;

    bucket = search_hashtable (variable_table, name);
    if (bucket != NULL) {
	Variable *var = (Variable *) (bucket->data);
	if (var->next != NULL) {
	    free_permanent_list (var->list);
	    bucket->data = (char *) var->next;
	    free_Variable (var);
	} else {
	    bucket = delete_hashtable (variable_table, name);
	    var = (Variable *) (bucket->data);
	    free_permanent_list (var->list);
	    free_Variable (var);
	    free_Bucket (bucket);
	}
    }
    return;
}


void init_variable_table ()
{
    static int initialized = 0;

    if (!initialized) {
	initialized = 1;
	variable_table = new_hashtable (0);
    }
}
